"""
The PSD Submodule
=================

The PSD submodule provides implementations of various particle size
distributions for the use in scattering calculations.

In addition to that, :code:`artssat.scattering.psd.arts` subpackage defines
the interface for PSDs in ARTS, while the :code:`artssat.scattering.psd.data`
subpackage provides functionality for the handling of PSD data.

"""
from artssat.scattering.psd.d14     import D14, D14N, D14MN
from artssat.scattering.psd.my05    import MY05
from artssat.scattering.psd.ab12    import AB12
from artssat.scattering.psd.binned      import Binned
from artssat.scattering.psd.fixed_shape import FixedShape



import numpy as np
import os

class LineCatalog:
    def __init__(self,
                 path):
        self.path = path

    def setup(self, workspace, sensors):
        workspace.ReadXML(workspace.abs_lines, self.path)

class Hitran(LineCatalog):
    def __init__(self, path = None):

        self.path = path
        if self.path is None:
            try:
                self.path = os.environ["HITRAN_PATH"]
            except:
                self.path = "HITRAN2012.par"

    def setup(self, workspace, sensors):

        f_max = 0.0
        f_min = np.finfo(np.float64).max
        for s in sensors:
            f_min = min(s.f_grid.min(), f_min)
            f_max = max(s.f_grid.max(), f_max)

        workspace.abs_linesReadFromHitran(self.path, f_min, f_max)

class Perrin(LineCatalog):
    def __init__(self, path = None):

        self.path = path
        if self.path is None:
            try:
                self.path = os.environ["PERRIN_PATH"]
            except:
                self.path = "spectroscopy/Perrin/"

    def setup(self, workspace, sensors):

        f_max = 0.0
        f_min = np.finfo(np.float64).max
        for s in sensors:
            f_min = min(s.f_grid.min(), f_min)
            f_max = max(s.f_grid.max(), f_max)

        df = f_max - f_min
        f_min = np.maximum(f_min - 0.5 * df, 0.0)
        f_max = f_max + 0.5 * df
        print(f_min, f_max)

        workspace.ReadSplitARTSCAT(basename = self.path,
                                   fmin = f_min,
                                   fmax = f_max)

class Aer(LineCatalog):
    def __init__(self, path):

        self.path = path

    def setup(self, workspace, sensors):
        try:
            workspace.ReadXML(workspace.abs_lines, self.path)
        except:
            workspace.ReadARTSCAT(filename = self.path)



"""
data_provider
-------------

The :code:`data_provider` module provides a base class for data provider
classes. This class provides utility function for the composition of data providers
as well as the overriding of get functions.
"""

import numpy as np
import weakref

################################################################################
# Data provider classes
################################################################################

class DataProviderBase:
    """
    The :code:`DataProviderBase` implements generic functionality for

    - The overriding of get methods with attributes
    - Composition of data providers

    When a data provider inherits from :code:`DataProviderBase`, the values
    returned from a :code:`get_<attribute>` can be overriding simply setting
    the attribute with the name :code:`<attribute>` of the object.

    The DataProviderBase also allows composing data provider by adding sub-
    providers to a data provider object. If a simulation queries a get_method
    from the data provider that the data provider doesn't provide, the query
    is forwarded to the subproviders in the order they were added. The get
    method from the first that provides such a method is returned.
    """
    def __init__(self):
        self._owner = None
        self.subproviders = []

    @property
    def owner(self):
        owner = self._owner()
        if owner:
            return owner
        else:
            raise ValueError("Parent data provider has been deleted.")

    @owner.setter
    def owner(self, owner):
        self._owner = weakref.ref(owner)

    def add(self, subprovider):
        """
        Add a subprovider to the data provider.

        Arguments:

            subprovider: The subprovider to add. Must inherit
                from :code:`DataProviderBase`.

        Raises:

            Exception: If :code: `subprovider` argument doesn´t inherit
            from :code:`DataProviderBase`.
        """
        if not isinstance(subprovider, DataProviderBase):
            raise Exception("Subprovider objects must inherit from DataProviderBase.")
        subprovider.owner = self
        self.subproviders += [subprovider]

    def __getattribute__(self, name):

        if not name[:4] == "get_":
            return object.__getattribute__(self, name)
        else:

            attribute_name = name[4:]

            #
            # Get name from self with priority for the attribute.
            #

            try:
                x = object.__getattribute__(self, attribute_name)

                def wrapper(*args, **kwargs):
                    return x
                return wrapper
            except:
                pass

            try:
                return object.__getattribute__(self, name)
            except:
                pass

            #
            # Get name from first list provider which has any of the
            # has either a fitting attribute or get_method.
            #

            for p in self.subproviders:

                try:
                    return getattr(p, name)
                except:
                    pass

            raise AttributeError("'{0}' object has no attribute '{1}'."\
                                 .format(type(self).__name__, name))

class CombinedProvider(DataProviderBase):
    """
    The :code:`CombinedProvider` allows the combination of multiple data
    providers. If a get method is requested from the data provider it is
    forwarded to the combined providers. The lookup is performed in the
    same order as the providers are provided to the constructor and the
    first hit is returned.
    """
    def __init__(self, *args):
        super().__init__()
        self.providers = list(args)

    def add(self, provider):
        """
        Add a provider to the combined data provider.

        Arguments:

            provider: The data provider to add. Must inherit
                from :code:`DataProviderBase`.

        Raises:

            Exception: If :code: `provider` argument doesn´t inherit
            from :code:`DataProviderBase`.
        """
        if not isinstance(subprovider, DataProviderBase):
            raise Exception("Data provider to add  objects must inherit from "
                            "DataProviderBase.")
        subprovider.owner = self
        self.providers += [provider]

    def __getattribute__(self, name):

        if not name[:4] == "get_":
            return object.__getattribute__(self, name)
        else:

            #
            # Get name from first list provider which has any of the
            # has either a fitting attribute or get_method.
            #

            for p in self.providers:

                try:
                    return getattr(p, name)
                except:
                    pass

            raise AttributeError("'{0}' object has no attribute '{1}'."\
                                 .format(type(self).__name__, name))

class Constant(DataProviderBase):
    """
    Returns a data provider that returns a constant value for a given
    quantity.
    """
    def __init__(self, name, value):
        """
        Create a data provider for quantity :code:`name` that
        returns the fixed value :code:`value`.

        Arguments:

            name(:code:`str`): Name of the quantity to provide. The
                created :class:`FunctorDataProvider` object will have
                a member function name :code:`get_<name>`.

            value: The value to return for the quantity :code:`quantity`.
        """
        super().__init__()
        self.value = value
        self.__dict__["get_" + name] = self.get

    def get(self, *args, **kwargs):
        return self.value

class FunctorDataProvider(DataProviderBase):
    """
    The FunctorDataProvider turns a function into a data provider that
    provides a get function for the results of the given function applied
    to a variable from the parent provider.
    """
    def __init__(self, name, variable, f):
        """
        Create a data provider for quantity :code:`name` that
        returns the result of function `f` applied to variable
        `name` from the parent provider.

        Arguments:

            name(:code:`str`): Name of the quantity to provide. The
                created :class:`FunctorDataProvider` object will have
                a member function name :code:`get_<name>`.

            variable(:code:`str`): The quantity to get from the parent
                data provider. When its get function is called the
                :code:`FunctorDataProvider` will call the :code:`get_<variable>`
                function of the parent data provider and the values as
                arguments to the function :code:`f`

            f(:code:`function`): Function to apply to the values of
                :code:`variable`. The results are returned as the values
                of the quantity :code:`name` by the :code:`FunctionDataProvider`
                object.
        """
        super().__init__()
        self.variable   = variable
        self.f          = f
        self.__dict__["get_" + name] = self.get

    def get(self, *args, **kwargs):
        try:
            f_get = getattr(self.owner, "get_" + self.variable)
            x = f_get(*args, **kwargs)
        except:
            raise Exception("Could not get variable {} from data provider."
                            .format(self.variable))

        y = self.f(x)
        return y



"""
Fixed shape PSD
===============

The fixed-shape PSD takes a fixed PSD shape and scales it with
a given mass density.
"""
from artssat.scattering.psd.arts.arts_psd import ArtsPSD
from artssat.scattering.psd.data.psd_data import PSDData, D_eq
from artssat.arts_object import arts_property
from artssat.arts_object import Dimension as dim
from artssat.arts_object import ArtsObject
from pyarts.workspace import Workspace, arts_agenda

ws = Workspace()

class FixedShape(ArtsPSD, ArtsObject):
    """
    Fixed shape particle size distribution.

    This PSD class takes a mass density as input and returns a given PSD shape
    scaled in the vertical dimension to match the provided mass density.
    """
    @arts_property("Vector")
    def x(self):
        return None

    @arts_property("Matrix")
    def shape(self):
        return None

    def __init__(self, x, data, size_parameter = D_eq(1000.0)):
        """
        Create PSD with shape given by :code:`x` and :code:`data`.

        Arguments:
            x(numpy.ndarray): 1D array containing the centers of the size
                bins of the provided PSD shape.

            data(numpy.ndarray): 1D array containing the PSD representing the
                shape of the PSD.

            size_parameter: :class:`SizeParameter` representing the size parameter
                over which the PSD is defined.
        """
        ArtsObject.__init__(self)
        ArtsPSD.__init__(self, self.psd.size_parameter)
        self.psd  = PSDData(x, data, size_parameter)
        self.x    = x
        self.data = data

        shape = self.psd.data.reshape((1, -1))
        self.shape = shape / self.psd.get_mass_density()
        self.size_parameter = self.psd.size_parameter

        self.name = "fixed_psd"

    @property
    def moment_names(self):
        return ["mass_density"]

    @property
    def pnd_call_agenda(self):
        @arts_agenda
        def pnd_call(ws):
            ws.Ignore(ws.pnd_agenda_input_t)
            ws.Ignore(ws.pnd_agenda_input)
            ws.Ignore(ws.pnd_agenda_input_names)
            ws.Ignore(ws.dpnd_data_dx_names)
            ws.Touch(ws.psd_data)
            ws.Touch(ws.dpsd_data_dx)

            md = ws.pnd_agenda_input.value.reshape(-1, 1)
            ws.pnd_size_grid = self.x
            ws.psd_size_grid = self.x
            ws.psd_data    = self.shape * md
            ws.psd_data_dx = self.shape

        return pnd_call

    def setup(self, ws, i):
        self.pbf_index = i

    def get_data(self, ws, i, *args, **kwargs):
        pass



"""
The ArtsPSD class
=================

"""

import numpy as np
import scipy as sp
from abc import abstractproperty
from artssat.scattering.psd.data.psd_data import SizeParameter, Area, D_eq,\
    D_max, Mass
from pyarts.workspace import arts_agenda

class ArtsPSD:
    r"""

    The interface for ARTS PSDs.

    In ARTS, A PSD for a given scattering property is defined by the
    corresponding agenda in the :code:`pnd_agenda_array`. The task
    of the :class:`ArtsPSD` class is to provide this agenda together
    with the names of the moments of the distribution.

    Attributes:

        size_parameter_names(dict): Dictionary that translates the
            :class:`SizeParameterEnum` into the corresponding name of
            the size parameter in ARTS.

    """

    properties = [("size_parameter", (), SizeParameter),
                  ("t_min", (), np.float),
                  ("t_max", (), np.float),
                  ("x_fit_start", (), np.float)]

    size_parameter_names = {D_eq : "dveq",
                            D_max : "dmax",
                            Mass : "mass",
                            Area : "area"}

    def __init__(self,
                 size_parameter,
                 t_min = 0.0,
                 t_max = 999.0):
        """
        Parameters:
            size_parameter(SizeParameter) :class:`SizeParameter` enum specifying the
                size parameter that is used by the PSD.
            t_min(numpy.float): ARTS parameter, minimum temperature for which PSD values
                will be produced.
            m_max(numpy.float): ARTS parameter, maximum temperature for which PSD values
                will be produced.
        """
        self.size_parameter = size_parameter
        self.t_min = t_min
        self.t_max = t_max

        self.x_fit_start = 100e-6

    #
    # Abstract properties
    #

    @abstractproperty
    def pnd_call_agenda(self):
        """
        The WSM call that is used to compute the PSD in ARTS.
        """
        pass

    @abstractproperty
    def moment_names(self):
        """
        List of strings that contains the names of the moments of the PSD.
        This is used in the ARTS :code:`particle_bulkrprop_names` and as
        name for the data requested from the data provider.
        """
        pass

    #
    # Generic PND agenda
    #

    @property
    def agenda(self):
        """
        The ARTS agenda representing the PSD. Should be used as :code:`pnd_agenda`
        in the ARTS :code:`pnd_agenda_array`.
        """
        size_parameter = ArtsPSD.size_parameter_names[type(self.size_parameter)]

        @arts_agenda
        def pnd_agenda(ws):
            ws.Ignore(ws.pnd_agenda_input_t)
            ws.Ignore(ws.pnd_agenda_input)
            ws.Ignore(ws.pnd_agenda_input_names)
            ws.Ignore(ws.dpnd_data_dx_names)
            ws.ScatSpeciesSizeMassInfo(species_index = ws.agenda_array_index,
                                       x_unit = size_parameter,
                                       x_fit_start = self.x_fit_start)
            ws.Copy(ws.psd_size_grid, ws.scat_species_x)
            ws.Copy(ws.pnd_size_grid, ws.scat_species_x)
            INCLUDE(self.pnd_call_agenda)
            ws.pndFromPsdBasic()

        return pnd_agenda



""" Dimensions

The Dimensions class keeps track of the different dimensions in
an ARTS atmosphere. The purpose of this class is to allow the inferral
of dimensions from the provided data, but at the same time to be able
to identify inconsistencies.

Attributes:

    p(Dimension): Dimension object representing the size of the
        ARTS pressure grid
    lat(Dimension) Dimension object representing the size of the
        latitude grid
    lon(Dimension) Dimension object representing the size of the
        longitude grid
    joker(Joker) Dimension object representing a dimensions that
        can take an arbitrary value.

"""

class Dimension:
    """
    Dimensions of data arrays in ARTS.

    The required dimensions of most array data used in an ARTS
    simulation are fixed by the dimensions of the atmosphere.
    The Dimension class provides a symbolic representation of
    these dimensions in an ARTS simulation, so that these can
    be inferred and checked for consistency.

    Dimension objects are used to define the dimensions of
    ARTS data and keep track of the different values inferred
    from user-provided data.

    """
    def __init__(self, name):
        """
        Create dimension with given name.

        Parameters:
            name(str): The name of the dimension, e.g. "Pressure grid"

        """
        self.name = name
        self.deductions = {}

    def add_deduction(self, ws, value, who):
        """
        If a value of a dimension becomes apparent from data it
        should be signaled to the :code:`Dimension` object. The
        dimension object keeps track of these "deductions" for
        each workspace separately.

        Parameters:

            ws(pyarts.workspace.Workspace): Workspace object
                for which the dimension was deduced.

            value(Int): The deduced dimenions

            who(str): The name of the variable from which this
                value was deduced.

        """
        if not ws in self.deductions:
            self.deductions[ws] = [(value, who)]
        else:
            self.deductions[ws] += [(value, who)]

    def check(self, ws, n, who):
        """
        Check given value for consistency.

        Parameters:

            ws(pyarts.workspace.Workspace): The workspace for which
                to check the deduction.

            n(int): The value which to check against previous deduced values.

        Returns:

            True if this is the first deduction or is previous deuctions agree
            with the current value. False otherwise.

        """

        if not ws in self.deductions:
            self.deductions[ws] = (n, who)
            return True
        else:
            (m, who) = self.deductions[ws]
            return m == n

    def get_value(self, ws):
        """
        Get the deduced value of the dimension.

        Parameters:

            ws(pyarts.workspace.Workspace) The workspace object for
                which the value of the dimension should be deduced.

        Raises:

            Exception: If no or inconsistent values have been deduced.

        """

        self.check(ws)

        ds = self.deductions[ws]
        if len(ds) > 0:
            return ds[0][0]
        else:
            raise Exception("No value for dimension of the {0} has been "
                            "deduced.")

class Joker(Dimension):
    """
    A dimension that can take an arbitrary value.

    """
    def __init__(self, name):
        """
        Create dimension with given name.

        Parameters:
            name(str): The name of the dimension, e.g. "Pressure grid"

        """
        self.name = name
        self.deductions = {}

    def add_deduction(self, ws, value, who):
        """
        This function does nothing.

        """
        pass

    def check(self, *args, **kwargs):
        """
        This function does nothing.

        """
        pass

    def get_value(self, *args, **kwargs):
        """
        Raises an exception because it doesn't make sense to try
        an get a value from a joker dimension.

        Raises:

            Exception: If no or inconsistent values have been deduced.

        """

        raise Exception("Can't get the value of a joker dimension.")

p     = Dimension("pressure grid")
lat   = Dimension("latitude grid")
lon   = Dimension("longitude grid")
atm   = Dimension("atmospheric dimension")
los   = Dimension("line of sight dimension")
joker = Joker("This dimension can take any value.")



"""
artssat.jacobian
----------------

The :code:`jacobian` module handles calculations of Jacobians in ARTS.
Functionality for computing Jacobians in ARTS is implemented through
three classes:

1. :class:`JacobianCalculation` handles the actual calculation of
   Jacobians and coordinates the interaction with the ARTS workspace.

2. :class:`JacobianQuantity` defines the general interface for quantities
   for which a Jacobian can be calculated and how to toggle the calculation
   of the Jacobian for a given quantity.

3. :class:`Jacobian` handles quantity-specific settings and results. This
   class must be defined for each Jacobian quantity separately.

Calculating Jacobians
=====================

To calculate the Jacobian of a quantity :code:`q` it suffices
to add it to the `jacobian` of the :class:`ArtsSimulation`:

::

    simulation.jacobian.add(q)

This will add the quantity :code:`q` to the quantities for which a Jacobian
should be computed. Moreover it will instantiate the :class:`Jacobian` class
of the quantity :code:`q` and set the :code:`q.jacobian` property of :code:`q`.

It is important to note that :code:`simulation.jacobian` and :code:`q.jacobian`
are of different types: :code:`simulation.jacobian` is of type
:code:`JacobianCalculation` and handles the calculation of the Jacobian for
multiple quantities whereas :code:`q.jacobian` holds the settings and results
of the Jacobian calculation for :code:`q`.

Reference
=========
"""

import scipy as sp
import scipy.interpolate
import numpy as np

from abc import ABCMeta, abstractmethod, abstractproperty

from artssat.sensor      import ActiveSensor, PassiveSensor
from artssat.arts_object import ArtsObject, arts_property
from artssat.arts_object import Dimension as dim

################################################################################
# Transformations
################################################################################

class Transformation(metaclass = ABCMeta):
    """
    Abstract base class for transformations of Jacobian quantities.

    ARTS allows the calculation of certain transformations of Jacobian
    quantities. In artssat, these transformations are represented by subclasses
    of the :class:`Transformation` class, which defines the general interface
    for transformations of Jacobian quantities.

    The interface consists of two functions:

    - :code:`setup`: When the :code:`setup` function is called the
      transformation object is expected to call the appropriate workspace
      function so that the transformation is added to the most recently added
      JacobianQuantity.

    - :code:`__call__`: Transformation objects should be callable and calling
      them should apply the transformation to a given numeric argument.
    """
    def __init__(self):
        pass

    @abstractmethod
    def setup(self, ws, data_provider, *args, **kwargs):
        pass

    @abstractmethod
    def __call__(self, x):
        pass

    @abstractmethod
    def invert(self, y):
        pass

class Log10(Transformation):
    """
    The decadal logarithm transformation $f(x) = \log_{10}(x)$.
    """
    def __init__(self, minimum = 1e-20):
        Transformation.__init__(self)
        self.minimum = minimum

    def setup(self, ws, data_provider, *args, **kwargs):
        ws.jacobianSetFuncTransformation(transformation_func = "log10")

    def __call__(self, x):
        return np.log10(np.maximum(x, self.minimum))

    def invert(self, y):
        return 10.0 ** y

class Log(Transformation):
    """
    The natural logarithm transformation $f(x) = \log_{10}(x)$.
    """
    def __init__(self):
        pass

    def setup(self, ws, data_provider, *args, **kwargs):
        ws.jacobianSetFuncTransformation(transformation_func = "log")

    def __call__(self, x):
        return np.log(x)

    def invert(self, y):
        return np.exp(1) ** y

class Atanh(Transformation):

    def __init__(self, z_min = 0.0, z_max = 1.0):
        Transformation.__init__(self)
        ArtsObject.__init__(self)

        self.z_min = z_min
        self.z_max = z_max

    @arts_property("Numeric")
    def z_min(self):
        return 0.0

    @arts_property("Numeric")
    def z_max(self):
        return 1.2

    def setup(self, ws, data_provider, *args, **kwargs):
        ws.jacobianSetFuncTransformation(transformation_func = "atanh",
                                         z_min = self.z_min,
                                         z_max = self.z_max)

    def __call__(self, x):
        x = np.minimum(x, 0.99 * self.z_max)
        x = np.maximum(x, self.z_min)
        return np.arctanh(2.0 * (x - self.z_min) / (self.z_max - self.z_min) - 1)

    def invert(self, y):
        return (np.tanh(y) + 1) * 0.5 * (self.z_max - self.z_min) + self.z_min

class Identity(Transformation):
    """
    The identity transformation $f(x) = x$.
    """
    def __init__(self):
        pass

    def setup(self, ws, data_provider, *args, **kwargs):
        pass

    def __call__(self, x):
        return x

    def invert(self, y):
        return y

class Composition(Transformation):
    """
    Composition of multiple transformations.

    The forward transformation is applied left to right
    as provided to the constructor.

    Arguments:
        *args: Sequence of transformations.
    """
    def __init__(self, *args):
        if not all([isinstance(a, Transformation) for a in args]):
            raise Exception("All provided transformation must implement the "
                            "abstract base class.")
        self.transformations = args


        if any([hasattr(t, "initialize") for t in self.transformations]):
            self.__dict__["initialize"] = self._initialize

    def _initialize(self, data_provider, *args, **kwargs):
        for t in self.transformations:
            if hasattr(t, "initialize"):
                t.initialize(data_provider, *args, **kwargs)

    def setup(self, ws, data_provider, *args, **kwargs):
        for t in self.transformations:
            t.setup(ws, data_provider, *args, **kwargs)

    def __call__(self, x):
        for t in self.transformations:
            x = t(x)
        return x

    def invert(self, y):
        for t in self.transformations[::-1]:
            y = t.invert(y)
        return y

################################################################################
# JacobianCalculation
################################################################################

class JacobianCalculation:
    """
    The :class:`JacobianCalculation` keeps track of the quantities for which
    to compute a Jacobian and coordinates their setup.
    """
    def __init__(self):
        """
        Initialize an empty Jacobian calculation.
        """
        self.jacobian_quantities = []

    def add(self, jq):
        """
        Add a quantity to list of quantities for which the Jacobian should
        be computed.

        Arguments:

            jq(:class:`JacobianQuantity`): Add Jacobian calculation for the given
            quantity. The quantity :code:`jq` must be an instance of the
            :class:`JacobianQuantity` abstract base class.
        """
        jq.jacobian = jq.jacobian_class(jq, len(self.jacobian_quantities))
        self.jacobian_quantities += [jq]

    def setup(self, ws, data_provider, *args, **kwargs):
        """
        Setup the Jacobian calculations on the given workspace.

        Initializes the definition of the Jacobian on the given workspace
        :code:`ws` by calling the :code:`jacobianInit` WSV on the given
        workspace. Calls the setup method for all quantities for which
        a Jacobian should be computed and finalizes the definition of
        the Jacobian by calling the :code:`jacobianClose` WSM.

        Arguments:

            ws(:code:`pyarts.workspace.Workspace`): Workspace object
            on which to setup the Jacobian calculation.

            data_provider: The data provider object providing the data for
            the simulation.

        """
        if not self.jacobian_quantities:
            ws.jacobianOff()
        else:
            ws.jacobianInit()
            for jq in self.jacobian_quantities:
                jq.jacobian.setup(ws)
                jq.transformation.setup(ws, data_provider, *args, **kwargs)
            ws.jacobianClose()


################################################################################
# JacobianQuantity
################################################################################

class JacobianQuantity(metaclass = ABCMeta):
    """
    Abstract interface for quantities for which a Jacobian can be computed.

    Quantities for which a Jacobian can be computed must expose a
    :code:`jacobian_class` which holds all quantity-specific WSM calls and
    settings required to compute its Jacobian.

    After a quantity has been added to the Jacobian quantities of a simulation,
    the :code:`jacobian_class` object representing the settings and results of
    the Jacobian calculation for this specific object can be accessed through
    its :code:`jacobian` property.
    """

    def __init__(self):
        self._transformation = None
        self._jacobian = None

    @abstractproperty
    def jacobian_class(self):
        """
        Return the class object that holds the actual implementation of the
        Jacobian calculation.
        """
        pass

    @property
    def jacobian(self):
        """
        The :code:`jacobian_class` object holding the quantity-specific settings
        and actual results of the Jacobian calculations for this quantity.
        """
        return self._jacobian

    @jacobian.setter
    def jacobian(self, j):
        if not isinstance(j, self.jacobian_class):
           #not isinstance(j, JacobianBase):
            raise ValueError("The jacobian property of a JacobianQuantity"\
                             " can only be set to an instance of the objects"\
                             "own jacobian_class.")
        else:
            self._jacobian = j

    @property
    def transformation(self):
        """
        The transformation to be applied to the retrieval quantity.
        """
        return self._transformation

    @transformation.setter
    def transformation(self, t):
        if not isinstance(t, Transformation):
            raise TypeError("The transformation of a retrieval quantity must"\
                            " be of type Transformation.")
        else:
            self._transformation = t

################################################################################
# JacobianBase
################################################################################

class JacobianBase(ArtsObject, metaclass = ABCMeta):
    """
    Abstract base class for the Jacobian classes that encapsulate the
    quantity-specific calls and settings of the Jacobian. This class
    muss be inherited by the :code:`jacobian_class` of each
    :code:`JacobianQuantity` object.
    """
    @arts_property("Vector", shape = (dim.Joker,))
    def p_grid(self):
        return np.zeros(0)

    @arts_property("Vector", shape = (dim.Joker,))
    def lat_grid(self):
        return np.zeros(0)

    @arts_property("Vector", shape = (dim.Joker,))
    def lon_grid(self):
        return np.zeros(0)

    def __init__(self, quantity, index):

        super().__init__()
        self.quantity = quantity
        self.index    = index

        if self.quantity.transformation is None:
            self.quantity.transformation = Identity()

    @property
    def name(self):
        return self.quantity.name

    @abstractmethod
    def setup(self, ws):
        """
        This method should call the :code:`jacobianAdd...` method corresponding to
        the quantity on the given workspace :code:`ws`.
        """
        pass

    def get_grids(self, ws):

        if self.p_grid.size == 0:
            g1 = ws.p_grid
        else:
            g1 = self.p_grid

        if self.lat_grid.size == 0:
            g2 = ws.lat_grid
        else:
            g2 = self.lat_grid

        if self.lon_grid.size == 0:
            g3 = ws.lon_grid
        else:
            g3 = self.lon_grid

        return {"g1" : g1, "g2" : g2, "g3" : g3}

    def interpolate_to_grids(self, x, grids):
        retrieval_grids = [self.p_grid,
                           self.lat_grid,
                           self.lon_grid]
        used_grids = []
        for i, g in enumerate(grids):
            if retrieval_grids[i].size == 0:
                used_grids += [g]
            else:
                used_grids += [retrieval_grids[i]]
        retrieval_grids = used_grids

        retrieval_grids_shape = [g.size for g in retrieval_grids]
        x = np.reshape(x, retrieval_grids_shape)

        x = x[::-1]
        retrieval_grids[0] = retrieval_grids[0][::-1]
        grids[0] = grids[0][::-1]

        interp = sp.interpolate.RegularGridInterpolator(retrieval_grids, x,
                                                        method = "linear",
                                                        bounds_error = False,
                                                        fill_value = np.nan)

        mesh_grids = np.meshgrid(grids)
        if len(mesh_grids) > 1:
            xi = np.transpose(np.stack(mesh_grids), axes = (0, -1))
        else:
            xi = mesh_grids[0].reshape(-1, 1)

        y  = interp(xi)

        inds = np.array(np.isnan(y))
        interp = sp.interpolate.RegularGridInterpolator(retrieval_grids, x,
                                                        method = "nearest",
                                                        bounds_error = False,
                                                        fill_value = None)
        y[inds] = interp(xi)[inds]

        grids_shape = [g.size for g in grids]
        y = np.reshape(y, grids_shape)[::-1]

        return y

    def get_data(self, ws, data_provider, *args, **kwargs):
        self.get_data_arts_properties(ws, data_provider, *args, **kwargs)



from copy import copy
import numpy as np

from pyarts.workspace     import Workspace
from artssat.sensor.sensor       import ActiveSensor, PassiveSensor
from artssat.scattering.solvers  import ScatteringSolver, RT4, Disort
from artssat.jacobian            import JacobianCalculation
from artssat.retrieval           import RetrievalCalculation
from artssat.io                  import OutputFile

################################################################################
# IPyParallel Cache
################################################################################

_ipyparallel_simulation = None

def setup_simulation_ipyparallel():
    global _ipyparallel_simulation
    _ipyparallel_simulation.setup()

def run_simulation_ipyparallel(args, **kwargs):
    global _ipyparallel_simulation
    _ipyparallel_simulation.run(*args, **kwargs)
    return _ipyparallel_simulation

class AsyncResults():
    def __init__(self, args):
        self.args = args
        self.done = {}
        self.failed = {}

    @property
    def average_time(self):
        t = np.array([self.done[k]["time"] for k in self.done])
        if len(t) > 0:
            return np.mean(t)
        else:
            return np.nan

    def __repr__(self):
        s = "Distributed ARTS simulation: {} tasks, {} completed, {} failed"
        s += "\n\t Avg. execution time: {}"
        return s.format(len(self.args),
                        len(self.done),
                        len(self.failed),
                        self.average_time)

################################################################################
# Arts Simulation
################################################################################

class ArtsSimulation:

    def __init__(self,
                 atmosphere = None,
                 data_provider = None,
                 sensors = [],
                 scattering_solver = RT4(nstreams = 12)):

        self._atmosphere        = atmosphere
        self._sensors           = sensors
        self._scattering_solver = scattering_solver
        self._data_provider     = data_provider
        self._setup             = False
        self._workspace         = None
        self.output_file        = None

        self.jacobian  = JacobianCalculation()
        self.retrieval = RetrievalCalculation()

        try:
            from mpi4py import MPI
            self.comm = MPI.COMM_WORLD
            size = self.comm.Get_size()
            self.parallel = size > 1
        except:
            self.parallel = False

        self.includes = ["general/general.arts",
                         "general/continua.arts",
                         "general/agendas.arts",
                         "general/planet_earth.arts"]

    #
    # Properties
    #

    @property
    def atmosphere(self):
        return self._atmosphere

    @property
    def sensors(self):
        return self._sensors

    @property
    def active_sensors(self):
        return [s for s in self._sensors if isinstance(s, ActiveSensor)]

    @property
    def passive_sensors(self):
        return [s for s in self._sensors if isinstance(s, PassiveSensor)]

    @property
    def data_provider(self):
        return self._data_provider

    @data_provider.setter
    def data_provider(self, dp):
        self._data_provider = dp

    @property
    def scattering_solver(self):
        return self._scattering_solver

    @scattering_solver.setter
    def scattering_solver(self, s):
        if isinstance(s, ScatteringSolver):
            self._scattering_solver = s
        else:
            raise ValueError("The scattering solver must be an instance of the "
                             "abstract ScatteringSolver class.")

    @property
    def workspace(self):
        return self._workspace

    #
    # Radiative transfer calculations
    #

    def setup(self, verbosity=0):

        self.verbosity  = verbosity
        self._workspace = Workspace(verbosity=verbosity,
                                    agenda_verbosity=verbosity)
        ws = self._workspace
        for include in self.includes:
            ws.execute_controlfile(include)

        ws.Copy(ws.ppath_agenda, ws.ppath_agenda__FollowSensorLosPath)
        ws.Copy(ws.ppath_step_agenda, ws.ppath_step_agenda__GeometricPath)
        ws.Copy(ws.iy_space_agenda, ws.iy_space_agenda__CosmicBackground)
        ws.Copy(ws.iy_surface_agenda, ws.iy_surface_agenda__UseSurfaceRtprop)
        ws.Copy(ws.iy_main_agenda, ws.iy_main_agenda__Emission)

        self.atmosphere.setup(ws, self.sensors)

        for s in self.sensors:
            s.setup(ws, self.atmosphere.scattering)

        self._setup = True

    def check_dimensions(self, f, name):
        s = f.shape

        err = "Provided atmospheric " + name + " field"
        err += " is inconsistent with the dimensions of the atmosphere."

        if len(s) != len(self.dimensions):
            raise Exception(err)
        if not all([i == j or j == 0 for i,j \
                    in zip(s, self.dimensions)]):
            raise Exception(err)

    def _run_forward_simulation(self, sensors = []):

        ws = self.workspace

        # Jacobian setup
        self.jacobian.setup(ws, self.data_provider, *self.args, **self.kwargs)

        # Run atmospheric checks
        self.atmosphere.run_checks(ws)

        i_active = []
        i_passive = []

        if len(sensors) == 0:
            sensors = self.sensors

        for i,s in enumerate(sensors):
            if isinstance(s, ActiveSensor):
                i_active += [i]
            if isinstance(s, PassiveSensor):
                i_passive += [i]


        if len(i_active) > 1:
            raise Exception("Currently at most one active sensor is "
                            "supported in a multi sensor observation.")

        i = 0
        y_index = 0

        if self.atmosphere.scattering:
            ws.pnd_fieldCalcFromParticleBulkProps()

        # Simulate active sensor
        if len(i_active) > 0:

            # TODO: Get around the need to fix this.
            ws.stokes_dim = s.stokes_dimension

            s = sensors[i_active[0]]

            if self.atmosphere.scattering:
                f = s.make_y_calc_function(append = False)
                f(ws)
            else:
                ws.y.value = s.y_min * np.ones(s.y_vector_length)
                ws.y_f = s.f_grid[0] * np.ones(s.y_vector_length)
                ws.y_pol = [0] * s.y_vector_length
                ws.y_pos = 0.0 * np.ones((s.y_vector_length,
                                          len(self.atmosphere.dimensions)))
                ws.y_geo = 0.0 * np.ones((s.y_vector_length, 5))
                ws.y_los = 0.0 * np.ones((s.y_vector_length,
                                          min(len(self.atmosphere.dimensions), 2)))

            i += 1
            if s.views > 1:
                s.y = np.copy(ws.y.value[y_index:].reshape((s.views, -1, s.stokes_dimension)))
            else:
                s.y = np.copy(ws.y.value[y_index:].reshape((-1, s.stokes_dimension)))
            y_index += ws.y.value.size


        # Simulate passive sensors
        for s in [sensors[i] for i in i_passive]:

            # TODO: Get around the need to fix this.
            ws.stokes_dim = s.stokes_dimension

            # Run scattering solver
            if self.atmosphere.scattering:
                f = self.scattering_solver.make_solver_call(self.atmosphere, s)
                f(ws)

            f = s.make_y_calc_function(append = i > 0,
                                       scattering = self.atmosphere.scattering)
            f(ws)

            if s.views > 1:
                s.y = np.copy(ws.y.value[y_index:].reshape((s.views, -1, s.stokes_dimension)))
            else:
                s.y = np.copy(ws.y.value[y_index:].reshape((-1, s.stokes_dimension)))
            y_index = ws.y.value.size

            i += 1


    def _run_retrieval(self, *args, **kwargs):

        ws = self.workspace
        scattering = len(self.atmosphere.scatterers) > 0

        self.retrieval.setup(self.workspace, self.sensors, self.scattering_solver,
                             scattering, self.data_provider, *args, **kwargs)


        ws.y = self.retrieval.y

        self.atmosphere.run_checks(ws)

        ws.covmat_seAddBlock(
            block = self.data_provider.get_observation_error_covariance(*args,
                                                                        **kwargs)
        )

        ws.OEM(**self.retrieval.settings)

    def run(self, *args, **kwargs):

        if not self._setup:
            raise Exception("setup() member function must be executed before"
                            " a simulation can be run.")

        self.args   = args
        self.kwargs = kwargs

        ws = self.workspace

        self.atmosphere.get_data(ws, self.data_provider, *args, **kwargs)
        for s in self.sensors:
            s.get_data(ws, self.data_provider, *args, **kwargs)

        # Run retrieval calculation
        if len(self.retrieval.retrieval_quantities) > 0:
            scattering = len(self.atmosphere.scatterers) > 0

            self.retrieval.run(self, *args, **kwargs)
        else:

            self._run_forward_simulation()
        return self

    def _run_ranges_mpi(self, ranges, *args, callback = None, **kwargs):
        from mpi4py import MPI

        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        size = comm.Get_size()

        r = ranges[0]
        n  = (r.stop - r.start) // r.step
        dn  = n // size

        if n < size:
            raise Exception("Need at least as many steps in outermost range as "
                            " processors. Otherwise deadlocks occur.")

        n0  = rank * dn
        rem = n % size
        if rank < rem:
            dn = dn + 1
        n0 += min(rank, rem)

        for i in range(r.start + n0, r.start + n0 + dn):
            self._run_ranges(ranges[1:], *args, i, **kwargs, callback = callback)

    def _run_ranges_ipyparallel(self,
                                ranges,
                                *args,
                                callback = None,
                                ipyparallel_client = None,
                                **kwargs):
        if ipyparallel_client is None:
            raise ValueError("To run simulations in parallel using ipyparallel,"
                             "the ipyparallel_client keyword argument must be"
                             " provided.")
        view = ipyparallel_client.direct_view()
        view.use_dill()

        view.block=True

        def run_setup(simulation):
            global _ipyparallel_simulation
            _ipyparallel_simulation = simulation
            _ipyparallel_simulation.setup()

        view.apply(run_setup, self)

        # Generate arguments
        def get_args(ranges):
            if len(ranges) == 1:
                return [(a,) for a in ranges[0]]
            else:
                return [(a, ) + seqs
                        for a in ranges[0]
                        for seqs in get_args(ranges[1:])]
        args = get_args(ranges)

        results = []
        view = ipyparallel_client.load_balanced_view()

        client = ipyparallel_client
        for i, arg in enumerate(args):
            results += [view.map(run_simulation_ipyparallel, [arg], block=False)]

        result_async = AsyncResults(args)

        def make_callback(arg, output_file):
            def done_callback(result):
                try:
                    simulation = result.get()[0]
                    if output_file:
                        output_file.store_results(simulation)
                    result_async.done[arg] = {"stdout" : result.stdout,
                                              "stderr" : result.stderr,
                                              "time" : (result.completed[0]
                                                        - result.started[0])}
                except Exception as e:
                    result_async.failed[arg] = {"exception" : e}
                del result
            return done_callback

        for arg, r in zip(args, results):
            r.add_done_callback(make_callback(arg, self.output_file))

        return result_async

    def _run_ranges(self, ranges, *args, callback = None, **kwargs):
        if len(ranges) == 0:
            self.run(*args, **kwargs)

            if not callback is None:
                callback(self)
            else:
                if not self.output_file is None:
                    self.store_results()
        else:
            r = ranges[0]
            for i in r:
                self._run_ranges(ranges[1:], *args, i, **kwargs)


    def run_ranges(self,
                   *args,
                   mpi = None,
                   ipyparallel_client = None,
                   callback = None,
                   **kwargs):

        if mpi is None:
            parallel = self.parallel
        else:
            if not ipyparallel_client is None:
                raise ValueError("Simulations can be run either using MPI or "
                                 " IPyParallel, not both. Therefore, only "
                                 "one of the mpi and ipyparallel_client keyword "
                                 "arguments can be given.")
            parallel = mpi

        ranges = list(args)

        if parallel:
            self._run_ranges_mpi(ranges, **kwargs, callback = callback)
        elif ipyparallel_client:
            return self._run_ranges_ipyparallel(ranges,
                                                **kwargs,
                                                callback = callback,
                                                ipyparallel_client=ipyparallel_client)
        else:
            self._run_ranges(ranges, **kwargs, callback = callback)

    def run_checks(self):
        self.atmosphere.run_checks(self.workspace)

    def initialize_output_file(self,
                               filename,
                               dimensions,
                               inputs=[],
                               mode = "w",
                               full_retrieval_output = True):
        self.output_file = OutputFile(filename,
                                      dimensions=dimensions,
                                      mode=mode,
                                      inputs=inputs,
                                      full_retrieval_output=full_retrieval_output)

    def store_results(self):
        if not self.output_file is None:
            self.output_file.store_results(self)
        else:
            raise Exception("The output file must be initialized before results"
                            " can be written to it.")

    def __getstate__(self):
        state = copy(self.__dict__)
        if "_workspace" in state.keys():
            state.pop("_workspace")
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._setup = False



"""
The model atmosphere
====================

"""

import numpy as np
from artssat.atmosphere.cloud_box import CloudBox
from artssat.jacobian import  JacobianBase
from artssat.retrieval import RetrievalBase, RetrievalQuantity
from artssat.atmosphere.catalogs import LineCatalog, Perrin

class TemperatureJacobian(JacobianBase):
    def __init__(self,
                 quantity,
                 index,
                 p_grid   = [],
                 lat_grid = [],
                 lon_grid = [],
                 hse      = "on"):
        super().__init__(quantity, index)
        self.p_grid   = p_grid
        self.lat_grid = lat_grid
        self.lon_grid = lon_grid
        self.hse      = hse

    def _make_setup_kwargs(self, ws):

        if self.p_grid.size == 0:
            g1 = ws.p_grid
        else:
            g1 = self.p_grid

        if self.lat_grid.size == 0:
            g2 = ws.lat_grid
        else:
            g2 = self.lat_grid

        if self.lon_grid.size == 0:
            g3 = ws.lon_grid
        else:
            g3 = self.lon_grid

        kwargs = {"g1" : g1, "g2" : g2, "g3" : g3,
                  "hse" : self.hse}

        return kwargs

    def setup(self, ws, data_provider, *args, **kwargs):
        kwargs = self._make_setup_kwargs(ws)
        ws.jacobianAddTemperature(**kwargs)

class TemperatureRetrieval(RetrievalBase, TemperatureJacobian):

    def __init__(self,
                 quantity,
                 index,
                 p_grid   = [],
                 lat_grid = [],
                 lon_grid = [],
                 hse = "on"):
        RetrievalBase.__init__(self)
        TemperatureJacobian.__init__(self, quantity, index,
                                     p_grid, lat_grid, lon_grid, hse)

    def add(self, ws):
        ws.retrievalAddTemperature(**self._make_setup_kwargs(ws))

class Temperature(RetrievalQuantity):


    def __init__(self, atmosphere):
        super().__init__()
        self.atmosphere = atmosphere

    def get_data(self, ws, data_provider, *args, **kwargs):
        t = data_provider.get_temperature(*args, **kwargs)
        self.atmosphere.__check_dimensions__(t, "temperature")
        ws.t_field = self.atmosphere.__reshape__(t)

    def set_from_x(self, ws, xa):
        x = self.transformation.invert(xa)
        self.t_field = x

    @property
    def name(self):
        return "temperature"

    @property
    def jacobian_class(self):
        return TemperatureJacobian

    @property
    def retrieval_class(self):
        return TemperatureRetrieval



class Atmosphere:

    def __init__(self,
                 dimensions,
                 absorbers = [],
                 scatterers = [],
                 surface = None,
                 catalog = None):

        self.__set_dimensions__(dimensions)
        self._required_data = [("p_grid", dimensions[:1], False),
                               ("temperature", dimensions, False),
                               ("altitude", dimensions, False),
                               ("surface_altitude", dimensions[1:], True)]

        self.absorbers  = absorbers
        self.scatterers = scatterers
        self.scattering = len(scatterers) > 0
        self._dimensions = dimensions
        self._cloud_box = CloudBox(n_dimensions = len(dimensions),
                                   scattering = self.scattering)



        self._surface_data_indices = []
        self._surface = surface
        self.temperature = Temperature(self)

        if not surface is None:
            nd = len(self._required_data)
            self._required_data += surface.required_data
            self.surface_data_indices = range(nd, len(self._required_data))

        self._catalog = catalog

    #
    # Dimensions
    #

    def __set_dimensions__(self, dimensions):
        if not type(dimensions) == tuple or not type(dimensions[0]) == int:
            raise Exception("Dimensions of atmosphere must be given as a tuple "
                            "of integers.")
        if not len(dimensions) in [1, 2, 3]:
            raise Exception("The number of dimensions of the atmosphere "
                            "must be 1, 2 or 3.")
        if not all([n >= 0 for n in dimensions]):
            raise Exception("The dimension tuple must contain only  positive "
                            "integers.")
        else:
            self._dimensions = dimensions

    @property
    def dimensions(self):
            return self._dimensions

    #
    # Absorbers
    #

    @property
    def absorbers(self):
        return self._absorbers

    @absorbers.setter
    def absorbers(self, absorbers):
        for a in absorbers:
            self.__dict__[a.name] = a
            self._required_data += [(a.name, self._dimensions, False)]
        self._absorbers = absorbers

    def add_absorber(self, absorber):
        self.__dict__[absorber.name] = absorber
        self._required_data += [(absorber.name, self._dimensions, False)]
        self._absorbers += absorber

    #
    # Cloud box
    #

    @property
    def cloud_box(self):
        return self._cloud_box

    #
    # Catalog
    #

    @property
    def catalog(self):
        """
        Line catalog from which to read absorption line data.
        """
        return self._catalog

    @catalog.setter
    def catalog(self, c):
        if isinstance(c, LineCatalog) or c is None:
            self._catalog = c
        else:
            raise ValueError("Line catalog must be of type LineCatalog.")

    #
    # Jacobian
    #

    def has_jacobian(self):
        for a in self.absorbers:
            if not a.jacobian is None:
                return True
        for b in self.scatterers:
            for m in b.moments:
                if not a.jacobian is None:
                    return True

    #
    # Scatterers
    #

    @property
    def scatterers(self):
        return self._scatterers

    @scatterers.setter
    def scatterers(self, scatterers):

        if not type(scatterers) is list:
            raise ValueError("The 'scatterers' property can only be set to a list.")

        for s in scatterers:
            self.__dict__[s.name] = s
            self._required_data += [(n, self._dimensions, False) \
                                    for n in s.moment_names]
        self._scatterers = scatterers
        self.scattering = True
        self._cloud_box = CloudBox(n_dimensions = len(self.dimensions),
                                   scattering = self.scattering)

    def add_scatterer(self, scatterer):
        self.__dict__[scatterer.name] = scatterer
        self._required_data += [(n, self._dimensions, False) \
                                for n in scatterer.moment_names]
        self._scatterers += [scatterer]
        self.scattering = True
        self._cloud_box = CloudBox(n_dimensions = len(self.dimensions),
                                   scattering = self.scattering)

    #
    # Surface
    #

    @property
    def surface(self):
        return self._surface

    @surface.setter
    def set_surface(self, s):
        if not self._surface is None:
            rd = [d for i, d in enumerate(self._required_data) \
                  if i not in self._required_data_indices]
            nd = len(rd)
            rd += surface.required_data

            self._required_data = rd
            self._requried_data_indices = range(nd, len(rd))
        self._surface = surface

    @property
    def required_data(self):
        return self._required_data

    #
    # Setup
    #

    def __setup_absorption__(self, ws, sensors):
        species = []
        lineshapes = []
        normalizations = []
        cutoffs = []

        for i, a in enumerate(self._absorbers):
            a.setup(ws, i)
            species += [a.get_tag_string()]
        ws.abs_speciesSet(species = species)

        # Set the line shape
        if not self.catalog is None:
            self.catalog.setup(ws, sensors)
            ws.abs_lines_per_speciesCreateFromLines()
            ws.abs_lines_per_speciesSetMirroring(option = "Same")
        else:
            for a in self._absorbers:
                if a.from_catalog:
                    raise Exception("Absorber {} has from_catalog set to true "
                                    "but no catalog is provided".format(a.name))
            ws.abs_lines_per_speciesSetEmpty()

        for i, a in enumerate(self._absorbers):
            tag = a.get_tag_string()
            cutoff = np.float32(a.cutoff)
            cutoff_type = a.cutoff_type
            #ws.abs_lines_per_speciesSetCutoffForSpecies(option = cutoff_type,
            #                                            value = cutoff,
            #                                            species_tag = tag)
            lineshape = a.lineshape
            ws.abs_lines_per_speciesSetLineShapeTypeForSpecies(option = lineshape,
                                                               species_tag = tag)

            normalization = a.normalization
            ws.abs_lines_per_speciesSetNormalizationForSpecies(option = normalization,
                                                               species_tag = tag)



        ws.Copy(ws.abs_xsec_agenda, ws.abs_xsec_agenda__noCIA)
        ws.Copy(ws.propmat_clearsky_agenda,
                ws.propmat_clearsky_agenda__OnTheFly)
        ws.lbl_checkedCalc()

    def __setup_scattering__(self, ws):
        ws.ScatSpeciesInit()
        pb_names = []
        for s in self._scatterers:
            s.setup(ws, len(pb_names))
            pb_names += s.moment_names
        ws.particle_bulkprop_names = pb_names

    def setup(self, ws, sensors):

        if len(self.dimensions) == 1:
            ws.AtmosphereSet1D()
        if len(self.dimensions) == 2:
            ws.AtmosphereSet2D()
        if len(self.dimensions) == 3:
            ws.AtmosphereSet3D()

        self.__setup_absorption__(ws, sensors)
        self.__setup_scattering__(ws)

        self.surface.setup(ws)
        self.cloud_box.setup(ws)

    def setup_jacobian(self, ws):

        for a in self.absorbers:
            a.setup_jacobian(ws)

        for s in self.scatterers:
            for m in s.moments:
                m.setup_jacobian(ws)

    #
    # Data
    #

    def __check_dimensions__(self, f, name):
        s = f.shape

        err = "Provided atmospheric " + name + " field"
        err += " is inconsistent with the dimensions of the atmosphere."

        if len(s) != len(self.dimensions):
            raise Exception(err)
        if not all([i == j or j == 0 for i,j \
                    in zip(s, self.dimensions)]):
            raise Exception(err)

    def __reshape__(self, f):
        s = [1, 1, 1]
        j = 0
        for i in range(len(self.dimensions)):
            if self.dimensions[0] > 0:
                s[i] = self.dimensions[i]
            else:
                s[i] = f.shape[i]
        return np.reshape(f, tuple(s))

    def __get_pressure__(self, ws, provider, *args, **kwargs):
        p = provider.get_pressure(*args, **kwargs).ravel()
        if self.dimensions[0] != 0 and p.size != self.dimensions[0]:
            raise Exception("Provided pressure grid is inconsistent with"
                            " dimensions of the atmosphere.")
        ws.p_grid = p

    def __get_altitude__(self, ws, provider, *args, **kwargs):
        dimensions = ws.t_field.value.shape
        z = provider.get_altitude(*args, **kwargs)
        self.__check_dimensions__(z, "altitude")
        z = self.__reshape__(z)
        if not z.shape == dimensions:
            raise Exception("Dimensions of altitude field inconsistent"
                            " with dimensions of temperature field.")
        ws.z_field = z

        # Surface altitude

        dimensions = ws.t_field.value.shape

        if hasattr(provider, "get_surface_altitude"):
            zs = provider.get_surface_altitude(*args, **kwargs)
            try:
                zs = zs.reshape(dimensions[1:])
                ws.z_surface = zs
            except:
                raise Exception("Shape " + str(zs.shape) + "of provided "
                                "surface altitude is inconsistent with "
                                "the horizontal dimensions of the "
                                "atmosphere " + str(dimensions) + ".")
        else:
            ws.z_surface = ws.z_field.value[0, :, :]

    def __get_absorbers__(self, ws, provider, *args, **kwargs):

        dimensions = ws.t_field.value.shape
        ws.vmr_field = np.zeros((len(self.absorbers),) + dimensions)

        for i, a in enumerate(self.absorbers):
            if a.retrieval is None:
                fname = "get_" + a.name
                f = provider.__getattribute__(fname)
                x = f(*args, **kwargs)
                self.__check_dimensions__(x, a.name)
                x = self.__reshape__(x)

                if not x.shape == dimensions:
                    raise Exception("Dimensions of " + a.name + " VMR field "
                                    "inconcistent with dimensions of temperature "
                                    "field.")
                ws.vmr_field.value[i, :, :, :] = x

        i = 0

        n_moments = sum([len(s.moment_names) for s in self.scatterers])
        ws.particle_bulkprop_field = np.zeros(((n_moments,)
                                               + ws.t_field.value.shape))

    def __get_scatterers__(self, ws, provider, *args, **kwargs):

        if not self.scatterers is None and len(self.scatterers) > 0:
            ws.cloudbox_on = 1
            ws.cloudboxSetFullAtm()

        dimensions = ws.t_field.value.shape

        for s in self.scatterers:
            s.get_data(ws, provider, *args, **kwargs)

    def get_data(self, ws, provider, *args, **kwargs):


        self.__get_pressure__(ws, provider, *args, **kwargs)
        self.temperature.get_data(ws, provider, *args, **kwargs)
        self.__get_altitude__(ws, provider, *args, **kwargs)
        self.__get_absorbers__(ws, provider, *args, **kwargs)
        self.__get_scatterers__(ws, provider, *args, **kwargs)

        self.cloud_box.get_data(ws, provider, *args, **kwargs)
        self.surface.get_data(ws, provider, *args, **kwargs)

    #
    # Checks
    #

    def run_checks(self, ws):

        ws.atmgeom_checkedCalc()
        ws.atmfields_checkedCalc(bad_partition_functions_ok = 1)
        ws.propmat_clearsky_agenda_checkedCalc()
        ws.propmat_clearsky_agenda_checkedCalc()
        ws.abs_xsec_agenda_checkedCalc()

        self.cloud_box.run_checks(ws)
        self.surface.run_checks(ws)


class Atmosphere1D(Atmosphere):

    def __init__(self,
                 absorbers = [],
                 scatterers = [],
                 surface = None,
                 levels = None,
                 catalog = None):
        if levels is None:
            dimensions = (0,)
        else:
            if not type(levels) == int:
                raise Exception("The number of levels of the 1D atmosphere "
                                "must be given by an integer.")
            else:
                dimensions = (level, )
        super().__init__(dimensions,
                         absorbers = absorbers,
                         scatterers = scatterers,
                         surface = surface,
                         catalog = catalog)



"""
The :code:`sensor` module bundles all ARTS functionality related to sensors
and provides a set of predefined sensor objects.
"""
from artssat.sensor.sensor import ICI, MWI, ActiveSensor, PassiveSensor, CloudSat








"""
artssat.retrieval.a_priori
--------------------------

The :code:`retrieval.a_priori` sub-module provides modular data provider
object that can be used to build a priori data providers.
"""
from artssat.data_provider import DataProviderBase
from artssat.sensor import ActiveSensor, PassiveSensor
from artssat.jacobian import Transformation
import numpy as np
import scipy as sp
import scipy.sparse


################################################################################
# Covariances
################################################################################

class Diagonal:
    def __init__(self,
                 diagonal,
                 mask = None,
                 mask_value = 1e-12):
        self.diagonal = np.array(diagonal)
        self.mask = mask
        self.mask_value = mask_value

    def get_covariance(self, data_provider, *args, **kwargs):

        if not self.mask is None:
            mask = np.logical_not(self.mask(data_provider, *args, **kwargs))
        else:
            mask = []

        if self.diagonal.size == 1:
            z = data_provider.get_altitude(*args, **kwargs)
            diagonal = self.diagonal * np.ones(z.size)
        else:
            diagonal = self.diagonal

        diagonal[mask] = self.mask_value
        diagonal = sp.sparse.diags(diagonal, format = "coo")
        return diagonal

class SpatialCorrelation:
    """
    Adds spatial correlation to a given covariance matrix.
    """
    def __init__(self,
                 covariance,
                 correlation_length,
                 correlation_type = "exp",
                 cutoff = 1e-12,
                 mask = None,
                 mask_value = 1e-12,
                 z = None):
        """
        Arguments:

            covariance: Covariance object providing the original covariance
                matrix to which to apply the spatial correlation.

            correlation_length(:code:`float`): Correlation length in meters.

            correlation_type(:code:`str`): Type of the correlation to apply.

            cutoff(:code:`float`): Threshold below which to set correlation
                coefficients to zero.
        """
        self.covariance         = covariance
        self.correlation_length = correlation_length
        self.correlation_type   = correlation_type
        self.cutoff = cutoff
        self.mask = mask
        self.mask_value = mask_value
        self.z = z

    def get_covariance(self, data_provider, *args, **kwargs):

        if self.z is None:
            z = data_provider.get_altitude(*args, **kwargs)
        else:
            z = self.z
        dz = np.abs(z.reshape(-1, 1) - z.reshape(1, -1))

        if self.correlation_type == "exp":
            corr = np.exp(- np.abs(dz / self.correlation_length) )
        elif self.correlation_type == "gauss":
            corr = np.exp(- (dz / self.correlation_length) ** 2)

        inds = corr < self.cutoff
        corr[inds] = 0.0

        covmat = self.covariance.get_covariance(data_provider, *args, **kwargs)
        if isinstance(covmat, sp.sparse.spmatrix):
            covmat = covmat.todense()
        diag = np.sqrt(covmat.diagonal())
        covmat = corr * np.array((diag.T * diag))

        if not self.mask is None:
            inds = np.logical_not(self.mask(data_provider, *args, **kwargs))
            inds2 = np.logical_or(inds.reshape(-1, 1), inds.reshape(1, -1))
            covmat[inds2] = 0.0
            covmat[inds, inds] = self.mask_value

        return np.asarray(covmat)

class Thikhonov:
    """
    Thikhonov regularization using second order finite differences.
    """
    def __init__(self,
                 scaling    = 1.0,
                 diagonal   = 0.0,
                 mask       = None,
                 mask_value = 1e12,
                 z_scaling  = True,
                 z_grid = None):
        """
        Arguments:
            scaling(:code:`np.float`): Scalar to scale the precision matrix with.

            mask: A mask object defining indices on the diagonal on the precision
                matrix to which to as :code:`mask_value`.

            mask_value(:code:`float`): Scalar to add to the diagonal of the precision
                matrix on locations defined by :code:`mask`.

            z_scaling(:code:`Bool`): Whether or not to scale matrix coefficients
                according to height differences between levels.
        """
        self.scaling    = scaling
        self.diagonal   = diagonal
        self.mask       = mask
        self.mask_value = mask_value
        self.z_scaling  = z_scaling
        self.z_grid = z_grid

    def get_covariance(self, data_provider, *args, **kwargs):
        precmat = self.get_precision(data_provider, *args, **kwargs)
        diag = precmat.diagonal()
        return sp.sparse.diags(1.0 / diag, format = "coo")

    def get_precision(self, data_provider, *args, **kwargs):

        if self.z_grid is None:
            z = data_provider.get_altitude(*args, **kwargs)
            z_old = None
        else:
            z = self.z_grid
            z_old = data_provider.get_altitude(*args, **kwargs)
        n = z.size

        du2 = np.ones(n - 2)

        du1 = -4.0 * np.ones(n - 1)
        du1[0]  = -2.0
        du1[-1] = -2.0

        dl1 = np.copy(du1)
        dl2 = np.copy(du2)

        d     = 6.0 * np.ones(n)
        d[:2]  = [1, 5]
        d[-2:] = [5, 1]

        if self.diagonal > 0.0:
            d += self.diagonal

        if not self.mask is None:
            mask = self.mask(data_provider, *args, **kwargs).astype(np.float)
            if not z_old is None:
                f = sp.interpolate.interp1d(z_old, mask,
                                            axis = 0,
                                            bounds_error = False,
                                            fill_value = (mask[0], mask[-1]))
                mask = f(z) > 0.5

            mask = np.logical_not(mask)
            du1[mask[:-1]] = 0
            du2[mask[:-2]] = 0
            dl1[mask[1:]]  = 0
            dl2[mask[2:]]  = 0
            d[mask] = self.mask_value


        precmat = sp.sparse.diags(diagonals = [du2, du1, d, dl1, dl2],
                                  offsets   = [2, 1, 0, -1, -2],
                                  format    = "coo")
        precmat *= self.scaling

        if self.z_scaling:
            zf = (np.diff(z) / np.diff(z).mean()) ** 2.0
            zf1 = np.zeros(z.shape)
            zf1[1:]  += zf
            zf1[:-1] += zf
            zf1[1:-1] *= 0.5
            precmat = sp.sparse.diags(diagonals = [zf1], offsets = [0], format = "coo") * precmat

        return precmat.tocoo()

################################################################################
# APrioriProviderBase
################################################################################

class APrioriProviderBase(DataProviderBase):
    def __init__(self,
                 name,
                 covariance):
        """
        Create :class:`DataProviderApriori` instance that will provide
        the value of the quantity :code:`name` from the owning data
        provider as a priori mean state.

        Arguments:

            name(:code:`name`): Name of the quantity that this a priori
                provider should propagate.

            covariance(:code:`numpy.array` or :code:`float`): Diagonal or
                full covariance matrix to be provided as covariance matrix
                for the retrieval quantity.

            spatial_correlation: Functor that is applied to the covariance
                matrix before this is returned.
        """

        super().__init__()
        xa_name = "get_" + name + "_xa"
        self.__dict__[xa_name] = self.get_xa

        if hasattr(self, "get_x0"):
            x0_name = "get_" + name + "_x0"
            self.__dict__[x0_name] = self.get_x0

        if hasattr(covariance, "get_covariance"):
            covariance_name = "get_" + name + "_covariance"
            self.__dict__[covariance_name] = self.get_covariance
        if hasattr(covariance, "get_precision"):
            precision_name = "get_" + name + "_precision"
            self.__dict__[precision_name] = self.get_precision

        if hasattr(self, "get_mask"):
            precision_name = "get_" + name + "_mask"
            self.__dict__[precision_name] = self.get_mask

        self.name = name
        self._covariance = covariance

    def get_covariance(self, *args, **kwargs):
        return self._covariance.get_covariance(self.owner, *args, **kwargs)

    def get_precision(self, *args, **kwargs):
        return self._covariance.get_precision(self.owner, *args, **kwargs)

################################################################################
# Masks
################################################################################

class And:
    """
    Creates a combined mask by applying logical and to a list
    of single masks.
    """
    def __init__(self, *args):
        self.masks = list(args)

    def __call__(self, data_provider, *args, **kwargs):
        """
        Arguments:

            data_provider: Data provider describing the atmospheric scenario.

            *args: Arguments to forward to data provider.

            **kwargs: Keyword arguments to forward to data_provider.
        """
        masks = []
        for m in self.masks:
            masks += [m(data_provider, *args, **kwargs)]

        m_and = masks[0]
        for m in masks[1:]:
            m_and = np.logical_and(m_and, m)

        return m_and

class TropopauseMask:
    """
    Returns a mask that is true only below the approximate height of the
    troposphere. The troposphere is detected as the first grid point
    where the lapse rate is negative and the temperature below 220.
    """
    def __init__(self):
        pass

    def __call__(self, data_provider, *args, **kwargs):
        """
        Arguments:

            data_provider: Data provider describing the atmospheric scenario.

            *args: Arguments to forward to data provider.

            **kwargs: Keyword arguments to forward to data_provider.
        """
        t     = data_provider.get_temperature(*args, **kwargs)
        t_avg = 0.5 * (t[1:] + t[:-1])
        lr    = - np.diff(t)

        tp = np.where(np.logical_and(lr < 0, t_avg < 220))[0]
        inds  = np.ones(t.size, dtype = np.bool)
        if len(tp > 0):
            i     = np.where(np.logical_and(lr < 0, t_avg < 220))[0][0]
            inds[i + 1 : inds.size] = False
        return inds

class FreezingLevel:
    def __init__(self,
                 lower_inclusive = False,
                 invert = False):
        self.lower_inclusive = lower_inclusive
        self.invert = invert

    def __call__(self, data_provider, *args, **kwargs):
        """
        Arguments:

            data_provider: Data provider describing the atmospheric scenario.

            *args: Arguments to forward to data provider.

            **kwargs: Keyword arguments to forward to data_provider.
        """
        t     = data_provider.get_temperature(*args, **kwargs)
        inds = np.where(t < 273.15)[0]
        if len(inds) > 0:
            i = inds[0]
        else:
            i = 0
        if self.lower_inclusive:
            i = max(i - 1, 0)
        inds = np.zeros(t.size, dtype = np.bool)
        inds[i:] = True

        if self.invert:
            inds = np.logical_not(inds)

        return inds

class TemperatureMask:
    """
    The temperature mask replaces values at grid points outside of the
    given temperature interval with another value.
    """
    def __init__(self,
                 lower_limit,
                 upper_limit,
                 lower_inclusive = False,
                 upper_inclusive = False):
        """
        Arguments:

            lower_limit(:code:`float`): The lower temperature limit
            upper_limit(:code:`float`): The upper temperature limit
            lower_inclusive: Whether or not to include highest grid
                point (in altitude) that does not lie in the half-open
                temperature interval between lower_limit and upper limit.
            lower_inclusive: Whether or not to include lowest grid
                point (in altitude) that does not lie in the half-open
                temperature interval between lower_limit and upper limit.

                point on lower limit.
            upper_inclusive: Whether or not to include adjacent grid
                point on upper limit.
        """
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.upper_inclusive = upper_inclusive
        self.lower_inclusive = lower_inclusive

    def __call__(self, data_provider, *args, **kwargs):
        t    = data_provider.get_temperature(*args, **kwargs)
        inds = np.logical_and(t.ravel() >= self.lower_limit,
                              t.ravel() <  self.upper_limit)
        if self.upper_inclusive:
            inds[1:] += inds[:-1]
        if self.lower_inclusive:
            inds[:-1] += inds[1:]
        return inds

class AltitudeMask:
    """
    The altitude mask replaces values at grid points outside of the
    given altitude interval with another value.
    """
    def __init__(self, lower_limit, upper_limit):
        """
        Arguments:

            lower_limit(:code:`float`): The lower altitude limit

            upper_limit(:code:`float`): The upper altitude limit

        """
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

    def __call__(self, data_provider, *args, **kwargs):
        z    = data_provider.get_altitude(*args, **kwargs)
        inds = np.logical_and(z.ravel() >= self.lower_limit,
                              z.ravel() <  self.upper_limit)
        return inds

################################################################################
# A priori providers
################################################################################

class DataProviderAPriori(APrioriProviderBase):
    """
    A priori provider that propagates an atmospheric quantity :code:`name`
    as a priori mean profile from the data provider.
    """

    def __init__(self,
                 name,
                 covariance):
        """
        Create :class:`DataProviderApriori` instance that will provide
        the value of the quantity :code:`name` from the owning data
        provider as a priori mean state.

        Arguments:

            name(:code:`name`): Name of the quantity that this a priori
                provider should propagate.

            covariance(:code:`numpy.array` or :code:`float`): Diagonal or
                full covariance matrix to be provided as covariance matrix
                for the retrieval quantity.

            spatial_correlation: Functor that is applied to the covariance
                matrix before this is returned.
        """

        super().__init__(name, covariance)

    def get_xa(self, *args, **kwargs):

        f_name = "get_" + self.name
        try:
            f = getattr(self.owner, f_name)
        except:
            raise Expetion("DataProviderApriori instance requires get method "
                           " {0} from its owning data provider.")

        x = f(*args, **kwargs)
        return x


class FixedAPriori(APrioriProviderBase):
    """
    Returns an a priori profile that does not depend on the atmospheric
    state.
    """
    def __init__(self,
                 name,
                 xa,
                 covariance,
                 mask = None,
                 mask_value = 1e-12):

        if not mask is None:
            if not hasattr(self, "get_mask"):
                self.__dict__["get_mask"] = self._get_mask

        super().__init__(name, covariance)
        self._xa   = np.array(xa)
        self.mask = mask
        self.mask_value = mask_value

    def _get_mask(self, data_provider, *args, **kwargs):
        if self.mask is None:
            xa = self._get_xa(data_provider, *args, **kwargs)
            mask = np.ones(xa.shape, dtype=np.bool)
        else:
            mask = self.mask(data_provider, *args, **kwargs)
        return mask

    def _get_xa(self, data_provider, *args, **kwargs):

        if self._xa.size == 1:
            z = data_provider.get_altitude(*args, **kwargs)
            xa = self._xa.ravel() * np.ones(z.size)
        else:
            xa = self._xa

        if not self.mask is None:
            mask = np.logical_not(self._get_mask(self, data_provider, *args, **kwargs))
            xa[mask] = self.mask_value

        return xa

    def get_xa(self, *args, **kwargs):
        return self._get_xa(self.owner, *args, **kwargs)

################################################################################
# Functional a priori
################################################################################

class FunctionalAPriori(APrioriProviderBase):
    """
    Returns an a priori profile that is a functional  transform
    of some variable.
    """
    def __init__(self,
                 name,
                 variable,
                 f,
                 covariance,
                 mask = None,
                 mask_value = 1e-12):

        if not mask is None:
            self.__dict__["get_mask"] = self._get_mask

        super().__init__(name, covariance)
        self.variable   = variable
        self.f          = f
        self.mask       = mask
        self.mask_value = mask_value

    def _get_mask(self, data_provider, *args, **kwargs):
        mask = self.mask(data_provider, *args, **kwargs)
        return mask.astype(np.float)

    def get_xa(self, *args, **kwargs):

        try:
            f_get = getattr(self.owner, "get_" + self.variable)
            x = f_get(*args, **kwargs)
        except:
            raise Exception("Could not get variable {} from data provider."
                            .format(self.variable))

        xa = self.f(x)

        if not self.mask is None:
            mask = np.logical_not(self.mask(self.owner, *args, **kwargs))
            xa[mask] = self.mask_value

        return xa

################################################################################
# Reference A Priori
################################################################################

class ReferenceAPriori(APrioriProviderBase):
    """
    Forwards value from data provider as a priori state. Useful for
    testing of retrieval implementation.
    """
    def __init__(self,
                 name,
                 covariance,
                 mask = None,
                 mask_value = 1e-12,
                 a_priori = None,
                 transformation = None,
                 variable = None):
        super().__init__(name, covariance)
        self.mask       = mask
        self.mask_value = mask_value
        self.a_priori = a_priori
        self.transformation = transformation

        if not variable is None:
            self.variable = variable
        else:
            self.variable = name

    def get_xa(self, *args, **kwargs):
        if not self.a_priori is None:
            self.a_priori.owner = self.owner
            return self.a_priori.get_xa(*args, **kwargs)
        else:
            f_get = getattr(self.owner, "get_" + self.variable)
            x = f_get(*args, **kwargs)
            if not self.transformation is None:
                x = self.transformation(x)
            return x

    def get_x0(self, *args, **kwargs):

        f_get = getattr(self.owner, "get_" + self.variable)
        x = f_get(*args, **kwargs)
        if not self.transformation is None:
            x = self.transformation(x)

        if not self.a_priori is None:
            if hasattr(self.a_priori, "mask") and not self.a_priori.mask is None:
                mask = np.logical_not(self.a_priori.mask(self.owner, *args, **kwargs))
                x[mask] = self.a_priori.mask_value

        return x

################################################################################
# Sensor a priori
################################################################################

class SensorNoiseAPriori(DataProviderBase):
    """
    Measurement error due to sensor noise.

    The :code:`SensorNoiseAPriori` class constructs a combined
    observation error covariance matrix from the noise characteristics
    of a list of sensors.

    The noise of particular sensors can be amplified by adding the
    scaling factor to the :code:`noise_scaling` attribute of the
    class.

    Attributes:

        noise_scaling(:code:`dict`): Dictionary mapping sensor names to
            noise scaling factors.
    """
    def __init__(self,
                 sensors):
        """
        Arguments:

            sensors(list of :code:`artssat.sensor.Sensor`): Sensors used
                in the observation for which to construct the observation
                error covariance matrix.

        """
        self.sensors = sensors
        self.noise_scaling = {}

    def get_observation_error_covariance(self, *args, **kwargs):
        m = 0

        stds = []

        for s in self.sensors:
            if isinstance(s, ActiveSensor):
                if s.name in self.noise_scaling:
                    c = self.noise_scaling[s.name]
                else:
                    c = 1.0
                stds += [c * s.nedt]

        for s in self.sensors:
            if isinstance(s, PassiveSensor):
                if s.name in self.noise_scaling:
                    c = self.noise_scaling[s.name]
                else:
                    c = 1.0
                stds += [c * s.nedt]

        sig = np.concatenate(stds).ravel()
        return sp.sparse.diags(sig ** 2.0, format = "coo")

################################################################################
# Reduced retrieval grids
################################################################################

class PiecewiseLinear(Transformation):
    def __init__(self, grid):
        self.grid = grid

    def initialize(self, data_provider, *args, **kwargs):
        old_grid, new_grid = self.grid._get_grids(*args, **kwargs)
        m = new_grid.size
        n = old_grid.size
        A = np.zeros((m, n))

        if m == 3:
            zl = new_grid[0]
            zr = new_grid[-1]
            A[0, old_grid < zl] = 1.0
            A[1, np.logical_and(old_grid >= zl, old_grid < zr)] = 1.0
            A[2, old_grid >= zr] = 1.0
        else:

            z = new_grid[0]
            zr = new_grid[1]
            conditions = [old_grid <= z, old_grid > z]
            values = [1.0, lambda x: np.maximum(1.0 - (x - z) / (zr - z), 0.0)]
            A[0, :] = np.piecewise(old_grid, conditions, values)

            for i in range(1, new_grid.size - 1):
                zl = new_grid[i - 1]
                z = new_grid[i]
                zr = new_grid[i + 1]
                conditions = [old_grid < z,
                            old_grid == z,
                            old_grid > z]
                values = [lambda x: np.maximum(1.0 - (z - x) / (z - zl), 0.0),
                        1.0,
                        lambda x: np.maximum(1.0 - (x - z) / (zr - z), 0.0)]
                A[i, :] = np.piecewise(old_grid, conditions, values)

            z = new_grid[-1]
            zl = new_grid[-2]
            conditions = [old_grid < z, old_grid >= z]
            values = [lambda x: np.maximum(1.0 - (z - x) / (z - zl), 0.0), 1.0]
            A[-1, :] = np.piecewise(old_grid, conditions, values)

        b = np.zeros(n)

        self.A = A
        self.b = b

    def setup(self, ws, data_provider, *args, **kwargs):
        self.initialize(data_provider, *args, **kwargs)
        ws.jacobianSetAffineTransformation(transformation_matrix = self.A,
                                           offset_vector = self.b)

    def __call__(self, x):
        A_ = A / np.sum(A, axis = -1, keepdims = True)
        return A_ @ (x - self.b)

    def invert(self, x):
        return self.A.T @ x + self.b


class ReducedVerticalGrid(APrioriProviderBase):

    def __init__(self,
                 a_priori,
                 grid,
                 quantity = "pressure",
                 covariance = None,
                 provide_retrieval_grid = True):

        if hasattr(a_priori, "mask"):
            if not a_priori.mask is None:
                if not hasattr(self, "get_mask"):
                    self.__dict__["get_mask"] = self._get_mask

        if covariance is None:
            super().__init__(a_priori.name, a_priori)
        else:
            super().__init__(a_priori.name, covariance)
        self.a_priori = a_priori
        self.new_grid = grid
        self.quantity = quantity
        self._covariance = covariance

        retrieval_p_name = "get_" + a_priori.name + "_p_grid"
        if provide_retrieval_grid:
            self.__dict__[retrieval_p_name] = self.get_retrieval_p_grid

    def _get_grids(self, *args, **kwargs):
        f_name = "get_" + self.quantity
        try:
            old_grid = getattr(self.owner, f_name)(*args, **kwargs)
        except:
            raise Exception("Data provider does not provide get function "
                            "for quantity {} required to determine original "
                            "size of retrieval grid."
                            .format(self.quantity))
        return old_grid, self.new_grid


    def _interpolate(self, y, *args, **kwargs):
        old_grid, new_grid = self._get_grids(*args, **kwargs)
        if self.quantity == "pressure":
            f = sp.interpolate.interp1d(old_grid[::-1], y[::-1],
                                        axis = 0,
                                        bounds_error = False,
                                        fill_value = (y[-1], y[0]))
            yi = f(new_grid[::-1])[::-1]
        else:
            f = sp.interpolate.interp1d(old_grid, y,
                                        axis = 0,
                                        bounds_error = False,
                                        fill_value = (y[0], y[-1]))
            yi = f(new_grid)
        return yi

    def _interpolate_matrix(self, y, *args, **kwargs):
        old_grid, new_grid = self._get_grids(*args, **kwargs)
        if self.quantity == "pressure":
            f = sp.interpolate.interp2d(old_grid[::-1], old_grid[::-1], y[::-1, ::-1],
                                        bounds_error = False, fill_value = np.nan)
            yi = f(new_grid[::-1], new_grid[::-1])[::-1, ::-1]
        else:
            f = sp.interpolate.interp2d(old_grid, old_grid, y,
                                        bounds_error = False)
            yi = f(new_grid, new_grid)

        if (yi.shape[0] > 1) and (yi.shape[1] > 1):
            # Check for degeneracy at boundaries.
            if np.all(yi[0, :] == yi[1, :]):
                yi[0, 0] = yi[1, 1]
                yi[0, 1:] = 0.0
                yi[1:, 0] = 0.0

            # Check for degeneracy at boundaries.
            if np.all(yi[-1, :] == yi[-2, :]):
                yi[-1, -1] = yi[-2, -2]
                yi[-1, :-1] = 0.0
                yi[:-1, -1] = 0.0
        return yi

    def _get_mask(self, data_provider, *args, **kwargs):
        mask = self.a_priori._get_mask(data_provider, *args, **kwargs)

        mask_i = self._interpolate(mask, *args, **kwargs)
        mask_i[mask_i > 0.0] = 1.0

        # Return if mask is false everywhere
        if np.all(np.logical_not(mask)):
            return mask_i


        old_grid, new_grid = self._get_grids(*args, **kwargs)
        i_0_old, i_1_old = np.where(mask)[0][[0, -1]]
        i_0_new, i_1_new = np.where(mask_i)[0][[0, -1]]

        if self.quantity == "pressure":
            if old_grid[i_0_old] > new_grid[i_0_new]:
                mask_i[max(i_0_new - 1, 0)] = 1.0
            if old_grid[i_1_old] < new_grid[i_1_new]:
                mask_i[min(i_1_new + 1, new_grid.size - 1)] = 1.0
        else:
            if old_grid[i_0_old] < new_grid[i_0_new]:
                mask_i[max(i_0_new - 1, 0)] = 1.0
            if old_grid[i_1_old] > new_grid[i_1_new]:
                mask_i[min(i_1_new + 1, new_grid.size - 1)] = 1.0

        return mask_i.astype(np.bool)

    def get_xa(self, *args, **kwargs):
        self.a_priori.owner = self.owner
        xa = self.a_priori.get_xa(*args, **kwargs)
        return self._interpolate(xa, *args, **kwargs)

    def get_x0(self, *args, **kwargs):
        self.a_priori.owner = self.owner
        if hasattr(self.a_priori, "get_x0"):
            x0 = self.a_priori.get_x0(*args, **kwargs)
        else:
            x0 = self.a_priori.get_xa(*args, **kwargs)
        return self._interpolate(x0, *args, **kwargs)

    def get_covariance(self, *args, **kwargs):
        if self._covariance is None:
            self.a_priori.owner = self.owner
            covmat = self.a_priori.get_covariance(*args, **kwargs)
            if isinstance(covmat, sp.sparse.spmatrix):
                covmat = covmat.todense()
            covmat = self._interpolate_matrix(covmat, *args, **kwargs)
            mask = self._get_mask(self.owner, *args, **kwargs)
            for i in np.where(np.logical_not(mask))[0]:
                covmat[i, i + 1 :] = 0.0
                covmat[i+1 :, i] = 0.0
            return covmat
        else:
            return self._covariance.get_covariance(self.owner, *args, **kwargs)

    def get_precision(self, *args, **kwargs):
        if self._covariance is None:
            self.a_priori.owner = self.owner
            precmat = self.a_priori.get_precision(*args, **kwargs)
            if isinstance(precmat, sp.sparse.spmatrix):
                precmat = precmat.todense()
            return self._interpolate_matrix(precmat, *args, **kwargs)
        else:
            return self._covariance.get_precision(self.owner, *args, **kwargs)

    def get_retrieval_p_grid(self, *args, **kwargs):
        if self.quantity == "pressure":
            return self._get_grids(*args, **kwargs)[1]
        else:
            p = self.owner.get_pressure(*args, **kwargs)
            return self._interpolate(p, *args, **kwargs)


class MaskedRegularGrid(ReducedVerticalGrid):

    def __init__(self,
                 a_priori,
                 n_points,
                 mask,
                 quantity = "pressure",
                 covariance = None,
                 provide_retrieval_grid = True,
                 transition = None):

        if not a_priori.mask is None:
            if not hasattr(self, "get_mask"):
                self.__dict__["get_mask"] = self._get_mask

        super().__init__(a_priori, None, quantity, covariance,
                         provide_retrieval_grid = provide_retrieval_grid)
        self.n_points = n_points
        self.mask = mask

        retrieval_p_name = "get_" + a_priori.name + "_p_grid"
        if provide_retrieval_grid:
            self.__dict__[retrieval_p_name] = self.get_retrieval_p_grid

        self.transition = transition

    def _get_mask(self, data_provider, *args, **kwargs):
        mask = np.ones((self.n_points + 2,))
        mask[1] = 0.0
        mask[-1] = 0.0
        return mask

    def _get_grids(self, *args, **kwargs):
        mask = self.mask
        f_name = "get_" + self.quantity
        try:
            old_grid = getattr(self.owner, f_name)(*args, **kwargs)
        except:
            raise Exception("Data provider does not provide get function "
                            "for quantity {} required to determine original "
                            "size of retrieval grid."
                            .format(self.quantity))

        mask = self.mask(self.owner, *args, **kwargs)

        if len(np.where(mask)[0]) > 0:
            i_first = np.where(mask)[0][0]
            i_last = np.where(mask)[0][-1]
        else:
            i_first = 0
            i_last  = len(mask) - 1

        n = min(self.n_points + 2, mask.sum() + 2)
        new_grid = np.zeros((n,))
        new_grid[1 : -1] = np.linspace(old_grid[i_first], old_grid[i_last],
                                       n - 2)

        if n == 2:
            new_grid[0] = old_grid[0]
            new_grid[1] = old_grid[2]
            return old_grid, new_grid

        # Left
        if self.transition is None:
            if i_first > 0:
                new_grid[-1] = old_grid[i_first - 1]
            else:
                new_grid[0] = 2 * new_grid[1] - new_grid[2]
        else:
            new_grid[0] = new_grid[1] - self.transition

        # Right
        if self.transition is None:
            if i_last < old_grid.size - 1:
                new_grid[-1] = old_grid[i_last + 1]
            else:
                new_grid[-1] = 2.0 * new_grid[-2] - new_grid[-3]
        else:
            new_grid[-1] = new_grid[-2] + self.transition

        return old_grid, new_grid



"""
Handling of PSD data
====================

The :code:`artssat.scattering.psd.data` subpackage provides the
:class:`PSDData` class for the handling of numeric PSD data. Numeric PSD data
here refers to a PSD that is not given in a parametrized form but instead by
an array of particle density values over a given size grid.
"""



from pyarts.workspace import WorkspaceVariable
from abc import ABCMeta, abstractmethod, abstractproperty
import numpy as np

class Jacobian:
    def __init__(self):
        self.method       = "analytical",
        self.perturbation = 1e-6,
        self.p_grid       = None,
        self.lat_grid     = None,
        self.lon_grid     = None

def extend_dimensions(x):
    shape = [1, 1, 1]
    if not type(x) == np.ndarray:
        x = np.array(x)
    for i, d in enumerate(x.shape):
        shape[i] = d
    return x.reshape(shape)

class AtmosphericQuantity(metaclass = ABCMeta):

    def __init__(self,
                 name,
                 dimensions,
                 jacobian = False):
        self._name = name

        self._data      = None
        self._wsv       = None
        self._wsv_index = None
        self._constant  = False

    #
    # Properties
    #

    @property
    def name(self):
        return self._name

    @property
    def dimensions(self):
        return self._dimensions

    @property
    def data(self):
        if type(self._data) == WorkspaceVariable:
            return self._data.value
        else:
            return self._data

    @data.setter
    def data(self, d):

        for i, d in enumerate(self.dimensions):
            if not d == 0 and not d == d.shape[i]:
                raise Exception("Dimensions of provided atmospheric {0}"
                                " field are inconsistent with the atmospheric"
                                " dimensions along dimension {1} ({2} vs. {3})."
                                .format(self.name, i, d, d.shape[i]))
        self._data = d
        self._constant = d

    @property
    def constant(self):
        return self._constant

    @constant.setter
    def constant(self, c):
        self._constant = c

    #
    # Methods
    #

    @abstractmethod
    def setup(self, ws, i):
        pass

    @abstractmethod
    def get_data(self, ws, provider, *args, **kwargs):
        dimensions = ws.t_field.value.shape
        f = provider.__getattribute__("get_" + self.name)
        x = f(*args, **kwargs)
        x = extend_dimensions(x)

        if not x.shape == dimensions:
            raise Exception("Shape of {0} field is inconsistent with "
                            "the dimensions of the atmosphere."
                            .format(self.name))

        self._wsv.value[self._wsv_index, :, :, :] = x



"""
PSD data
========

Particle size distribution (PSD) data consists of a range of particle
densities given w.r.t to a corresponding grid of particle sizes. This
module prvodies functionality for the handling of such PSD data.

Class reference
---------------
"""
import numpy as np

################################################################################
# Size parameter classes
################################################################################

class SizeParameter:
    """
    General representation of a size parameter of a PSD. The size parameter
    is represented by its mass-size relation, which gives the mass :math:`m`
    as a function of the size parameter :math:`X`:

    .. math::
        m(X) = a \cdot X^b

    Attributes:

        a(numpy.float): The :math:`a` factor of the mass-size relation

        b(numpy.float): The :math:`b` exponent of the mass-size-relation
    """

    def __eq__(self, other):
        return self.a == other.a and self.b == other.b

    def __init__(self, a, b):
        """
        Create size parameter with given :math:`a` and :math:`b` parameters.
        """
        self.a = a
        self.b = b

    def convert(self, src, x, y):
        """
        Conversion of PSD data given over one size parameter to another.

        Parameters:

            src(SizeParameter): The size parameter in which the PSD data is given.

            x(numpy.ndarray): The size grid over which the PSD data is given.
            Must be broadcastable into the shape of :code:`y`.

            y(numpy.ndarray): The PSD data with the last axis corresponding to
            the size grid :code:`x`.

        Returns:

            :code:`(x, y)`: Tuple containing the transformed size grid :code:`x`
            and PSD data :code:`y`
        """
        x = (src.a / self.a) ** (1.0 / self.b) * x ** (src.b / self.b)
        y = y * (self.a / src.a) ** (1.0 / src.b) * (self.b / src.b) \
            * x ** (self.b / src.b - 1.0)
        return x, y

    def get_mass_density(self, x, y):
        """
        Compute the mass of the PSD for given PSD data.

        Parameters:

            x(numpy.ndarray): The size grid over which the PSD data is given.
            Must be broadcastable into the shape of :code:`y`.

            y(numpy.ndarray): The PSD data with the last axis corresponding to
            the size grid :code:`x`.

        """
        return np.trapz(self.a * x ** self.b * y, x = x, axis = -1)

class Area(SizeParameter):
    def __ini__(self, a, b):
        super().__init__(a, b)

class D_eq(SizeParameter):

    def __init__(self, rho):
        self.rho = rho
        super().__init__(self.rho * np.pi / 6.0, 3.0)

class D_max(SizeParameter):
    def __init__(self, a, b):
        super().__init__(a, b)

class Mass(SizeParameter):
    def __init__(self):
        super().__init__(1.0, 1.0)

################################################################################
# PSD Data
################################################################################

class PSDData:
    """
    The :code:`PSDData` class handles PSD data given as a discrete distribution
    over a given size grid.

    The PSD data is stored as a :code:`numpy.ndarray` of which the last dimension
    is assumed to correspond to the size parameter.

    Attributes:

        data(numpy.array): The discrete PSD data.

        x(numpy.array): The size grid corresponding to the last dimension in
        :code:`data`

        size_parameter(SizeParameter): The size parameter used to represent
        the PSD.
    """
    def __init__(self, x, data, size_parameter):
        """
        Create a PSDData object from data.

        The particle number density values should be given as a 2D array
        :code:`data`, with the first dimenions corresponding to PSDs from
        different volume elements and the first dimension corresponding to
        the size paramter.

        Parameters:

            x(numpy.array): The size grid used for :code:`data`

            data(numpy.array): Array containing particle number density values

            size_paramter(SizeParameter): Enum representing the type of size
                parameter used.

        """
        x_shape = (len(data.shape) - 1) * (1,) + (-1,)
        x = x.reshape(x_shape)

        if data.shape[-1] != x.shape[-1]:
            raise Exception("Size grid 'x' and PSD data 'y' must have the same"\
                            " number of elements along the last dimension.")


        self.data = data
        self.x = x

        if not isinstance(size_parameter, SizeParameter):
            raise Exception("Provided size_parameter is not a valid value of "
                            "of type SizeParameter.")
        self.size_parameter = size_parameter

    def get_moment(self, p, reference_size_parameter = None):
        """
        Compute the :math:`p` th moment :math:M(p) of the PSD data:

        .. math::
            M(p) = \int_0^\infty X^P \frac{dN}{dX}(X)\:dX

        The physical significance of a moment of a PSD depends on the size
        parameter. So in general, the moments of the same PSD given w.r.t.
        different size parameters differ. If :code: `reference_size_parameter`
        argument is given then the computed moment will correspond to the
        Moment of the PSD w.r.t. to the given size parameter.

        Parameters:

            p(numpy.float): The moment which to compute.

            reference_size_parameter(:class: `SizeParameter`): Size parameter
            with respect to which the moment should be computed.

        Return:
            The :math:`p` th moment of the particle size distribution.

        """
        if not reference_size_parameter is None:
            a1 = self.size_parameter.a
            b1 = self.size_parameter.b
            a2 = reference_size_parameter.a
            b2 = reference_size_parameter.b

            c = (a1 / a2) ** (p / b2)
            p = p * b1 / b2
        else:
            c = 1.0

        return c * np.trapz(self.data * self.x ** p, x = self.x)

    def __add__(self, other):
        """
        Adds the data from two particle size distributions. For this the PSDs
        must be given w.r.t to the same size parameter type and grid.

        Parameters:

            other(:code:`PSDData`): The PSD to add.

        Returns:

            :code:`PSDData` object representing the sum of the two PSDs.

        """
        # TODO: Make an abstract base class for PSDs
        if hasattr(other, "evaluate"):
            x2, _ = other.size_parameter.convert(self.size_parameter,
                                                 self.x,
                                                 self.data)
            other = other.evaluate(self.x2)
            other.change_size_parameter(self.size_parameter)

        if (not self.size_parameter.a == other.size_parameter.a) or \
           (not self.size_parameter.b == other.size_parameter.b):
           raise Exception("Addition of PSD data is only defined for PSDs"
                           " given w.r.t. the same size parameter.")

        if not np.all(np.isclose(self.x, other.x)):
            raise Exception("Addition of PSD data is only defined for PSDs"
                            " defined over the same size grid.")

        size_parameter = self.size_parameter
        y = self.data + other.data
        x = self.x

        return PSDData(x, y, size_parameter)

    def get_number_density(self):
        """
        Computes the total particle number density (:math:`0` th moment).

        Returns:
            The particle number density for each volume element.

        """
        return self.get_moment(p = 0)

    def get_mass_density(self):
        """
        Computes the particle mass density corresponding to the PSD data.

        Returns:
            Array containing the particle mass density for each of the
            bulk volumes described by the PSD data
        """
        return self.size_parameter.get_mass_density(self.x, self.data)

    def change_size_parameter(self, size_parameter):
        """
        Change the size parameter of the data.

        Parameters:
            size_parameter(SizeParameter): The size parameter to convert the
                data to.

            rho(np.float): Particle density.

        """

        self.x, self.data = size_parameter.convert(self.size_parameter,
                                                   self.x, self.data)
        self.size_parameter = size_parameter



class CloudBox:
    def __init__(self, n_dimensions, scattering = True):

        self._scattering = scattering
        self._adaptive = None
        self._vertical_limits = None
        self._vertical_limits_type = None
        self._latitude_limits = None
        self._longitude_limits = None

        self._checked = False

    @property
    def limits(self):
        return self._limits

    #
    # Adaptive
    #

    @property
    def adaptive(self):
        return self._adaptive

    @adaptive.setter
    def set_adaptive(self, a):
        self._altitude_limits = None
        self._pressure_limits = None
        self._adaptive = a

    #
    # Altitude limits
    #

    @property
    def altitude_limits(self):
        if self._vertical_limits is None \
           or self._vertical_limits_type != "altitude":
            raise Exception("No altitude limits have been set.")
        return self._vertical_limits

    @altitude_limits.setter
    def set_altitude_limits(self, z1, z2):
        self._adaptive = False
        self._vertical_limits_type = "altitude"

        if not z1 < z2:
            raise Exception("The cloudbox limits pressure limits must "
                            "satisfy z1 > z2.")
        self._vertical_limits = (z1, z2)

    #
    # Pressure limits
    #

    @property
    def pressure_limits(self):
        if self._vertical_limits is None \
           or self._vertical_limits_type != "pressure":
            raise Exception("No pressure limits have been set.")
        return self._pressure_limits

    @pressure_limits.setter
    def set_pressure_limits(self, p1, p2):

        self._adaptive = False
        self._vertical_limits_type == "pressure"

        if not p1 > p2:
            raise Exception("The cloudbox limits pressure limits must "
                            "satisfy p1 > p2.")
        self._pressure_limits = (p1, p2)

    #
    # Latitude limits
    #

    @property
    def latitude_limits(self):
        if self._latitude_limits is None:
            raise Exception("No latitude limits have been set.")
        return self._latituce_limits

    @latitude_limits.setter
    def set_latitude_limits(self, l1, l2):

        if not l1 < l2:
            raise Exception("The latitude limits must satisfy l1 < l2.")

        self._latitude_limits = (l1, l2)

    @property
    def checked(self):
        return self._checked

    def setup(self, ws):
        ws.Copy(ws.iy_cloudbox_agenda, ws.iy_cloudbox_agenda__LinInterpField)

    def get_data(self, ws, provider, *args, **kwargs):

        if not self._scattering:
            ws.jacobianOff()
            ws.cloudboxOff()
            ws.cloudbox_checked = 1
            return

        if not self._adaptive and self._vertical_limits is None:
            ws.cloudboxSetFullAtm()
            return

        if self._adaptive:
            ws.cloudboxSetAutomatically(particle_field =
                                        ws.particle_bulkprop_field)
            return

        v1 = self._vertical_limits[0]
        v2 = self._vertical_limits[1]

        lat_1 = 1.0
        lat_2 = 1.0

        if not self._latitude_limits is None:
            lat_1 = self._latitude_limits[0]
            lat_2 = self._latitude_limits[1]
        elif self._n_dimensions > 1:
            lat_1 = ws.lat_grid.value[0]
            lat_2 = ws.lat_grid.value[-1]

        lon_1 = 1.0
        lon_2 = 1.0

        if not self._longitude_limits is None:
            lon_1 = self._latitude_limits[0]
            lon_2 = self._latitude_limits[1]
        elif self._n_dimensions > 2:
            lon_1 = ws.lon_grid.value[0]
            lon_2 = ws.lon_grid.value[-1]

        if self._vertical_limits_type == "pressure":
            ws.cloudboxSetManually(p1 = v1,
                                   p2 = v2,
                                   lat1 = lat1,
                                   lat2 = lat2,
                                   lon1 = lon1,
                                   lon2 = lon2)

        if self._vertical_limits_type == "altitude":
            ws.cloudboxSetManuallyAltitude(z1 = v1,
                                           z2 = v2,
                                           lat1 = lat1,
                                           lat2 = lat2,
                                           lon1 = lon1,
                                           lon2 = lon2)
            ws._checked = False

    def run_checks(self, ws):

        if not self._scattering:
            return None

        ws.pnd_fieldCalcFromParticleBulkProps()
        ws.cloudbox_checkedCalc()




"""
Surface
=======

"""
import numpy as np
from pyarts.workspace import Workspace, arts_agenda

ws = Workspace()
ws.NumericCreate("surface_temperature")
ws.NumericCreate("surface_salinity")
ws.NumericCreate("surface_windspeed")

from artssat.atmosphere.surface.surface import Tessem, Telsem, CombinedSurface



"""
The surface sub-module provides implementations of the different surface models
that are available in ARTS.
"""
from abc import abstractmethod, abstractproperty
import numpy as np

from artssat.arts_object import ArtsObject, arts_property, Dimension
from pyarts.workspace import Workspace, arts_agenda
from pyarts.workspace.variables import (WorkspaceVariable, workspace_variables)
wsv = workspace_variables

ws = Workspace(verbosity = 0)


class Surface:
    """
    Abstract base class for surfaces.

    This class defines the general interfaces for surface models in artssat.
    """

    @abstractproperty
    def required_data(self):
        pass

    @abstractmethod
    def setup(self, ws):
        """
        Setup the surface model in the given workspace.

        This method should be used to setup the surface model so that it can be
        used in a simulation.

        Args:
            ws: pyarts.workspace.Workspace object on which the simulation will
                be performed.
        """
        pass

    @abstractmethod
    def get_data(self, ws, data_provider, *args, **kwargs):
        """
        The method should get required data from the data provider and set required
        field in the workspace.

        Args:
            ws: pyarts.workspace.Workspace object on which the simulation will
                be performed.
            provider: The data provider for the simulations.
            *args: Arguments to be passed to the data provider.
            **kwargs: Keyword arguments to be passed to the data provider.

        """
        pass

    @abstractmethod
    def run_checks(self, workspace):
        """
        This method should run the tests that are required before the model
        can be used in a simulation.

        Args:
            workspace: The pyarts.workspace.Workspace in which the simulations are
                performed.
        """
        pass

################################################################################
# Tessem
################################################################################

#TODO: Need elegant system for on demand WSVs.
telsem_salinity = ws.create_variable("Numeric", None)
telsem_windspeed = ws.create_variable("Numeric", None)

class Tessem(Surface,
             ArtsObject):
    """
    This class represents the " Tool to Estimate Sea‐Surface Emissivity from
    Microwaves to sub‐Millimeter waves" (TESSEM). It is a parametrization for
    the emissivity of sea surfaces that uses two neural networks to predict the
    emissivity.
    """

    def __init__(self,
                 tessem_net_h = "testdata/tessem_sav_net_V.txt",
                 tessem_net_v = "testdata/tessem_sav_net_V.txt",
                 salinity = 0.034,
                 wind_speed = 0.0):
        """
        Args:
            tessem_net_h(:code:`str`): Path to the network to use for
                h-polarization.
            tessem_net_v(:code:`str`): Path to the network to use for
                v-polarization.
            salinity (float): The salinity to assume for the sea surface.
            wind_speed (float): Wind speed if it should be set to a fixed
                value.
        """
        Surface.__init__(self)
        ArtsObject.__init__(self)

        self.tessem_net_h = tessem_net_h
        self.tessem_net_v = tessem_net_v

    @property
    def required_data(self):
        return [("surface_temperature", ("n_lat", "n_lon"), False),
                ("surface_salinity", (1,), True),
                ("surface_salinity", ("n_lat", "n_lon"), True),
                ("surface_wind_speed", (1,), True),
                ("surface_wind_speed", ("n_lat", "n_lon"), True)]

    @arts_property(group="Matrix",
                   shape=(Dimension.Lat, Dimension.Lon),
                   wsv=wsv["t_surface"])
    def surface_temperature(self):
        return 280.0

    @arts_property(group="Numeric",
                   wsv=telsem_salinity)
    def salinity(self):
        return 0.035

    @arts_property(group="Numeric",
                   wsv=telsem_windspeed)
    def surface_wind_speed(self):
        return 1.0

    @property
    def surface_agenda(self):

        @arts_agenda
        def surface_rtprop_agenda_tessem(ws):
            ws.specular_losCalc()
            ws.InterpSurfaceFieldToPosition(out = wsv["surface_skin_t"],
                                            field = wsv["t_surface"])
            ws.surfaceTessem(salinity = telsem_salinity,
                                wind_speed = telsem_windspeed)
        return surface_rtprop_agenda_tessem

    def setup(self, ws):
        ws.TessemNNReadAscii(ws.tessem_neth, self.tessem_net_h)
        ws.TessemNNReadAscii(ws.tessem_netv, self.tessem_net_v)
        ws.Copy(ws.surface_rtprop_agenda, self.surface_agenda)

    def get_data(self, ws, data_provider, *args, **kwargs):
        self.get_data_arts_properties(ws, data_provider, *args, **kwargs)

    def run_checks(self, ws):
        pass

################################################################################
# Telsem
################################################################################

class Telsem(Surface,
             ArtsObject):
    """
    This class implements the Tool for estimating surface emissivities 2
    (TELSEM2). TELSEM2 is a microwave atlas covering microwave and
    sub-millimeter frequencies. It comes in the form of 12 microwave atlases,
    one for each month that contain microwave emissivities of land-surfaces on
    an equal-area grid.
    """
    def __init__(self,
                 atlas_directory,
                 month=6):
        """
        Args:
            atlas_directory: Directory containing the Telsem atlases
            month: The month of the atlas to use.
        """
        Surface.__init__(self)
        ArtsObject.__init__(self)

        self.atlas_directory = atlas_directory
        self.month = month

        self.filename_pattern = "ssmi_mean_emis_climato_@MM@_cov_interpol_M2"
        self.r_min = 0.0
        self.r_max = 1.0
        self.d_max = -1.0

    @property
    def required_data(self):
        return [("temperature", ("n_lat", "n_lon"), False)]

    @arts_property(group="Vector",
                   shape=(Dimension.Lat,),
                   wsv=wsv["lat_true"])
    def latitude(self):
        return None

    @arts_property(group="Vector",
                  shape=(Dimension.Lon,),
                  wsv=wsv["lon_true"])
    def longitude(self):
        return None

    @arts_property(group="Matrix",
                   shape=(Dimension.Lat, Dimension.Lon),
                   wsv=wsv["t_surface"])
    def surface_temperature(self):
        return None

    @property
    def surface_agenda(self):

        @arts_agenda
        def surface_rtprop_agenda_telsem(ws):
            ws.specular_losCalc()
            ws.InterpSurfaceFieldToPosition(out = wsv["surface_skin_t"],
                                            field = wsv["t_surface"])
            ws.surfaceTelsem(atlas=ws.telsem_atlas,
                             r_min=self.r_min,
                             r_max=self.r_max,
                             d_max=self.d_max)
        return surface_rtprop_agenda_telsem

    def setup(self, ws):
        ws.TelsemAtlasCreate("telsem_atlas")
        ws.telsem_atlasesReadAscii(directory=self.atlas_directory,
                                   filename_pattern=self.filename_pattern)
        ws.Extract(ws.telsem_atlas, wsv["telsem_atlases"], self.month)
        ws.Copy(ws.surface_rtprop_agenda, self.surface_agenda)

    def get_data(self, ws, data_provider, *args, **kwargs):
        self.get_data_arts_properties(ws, data_provider, *args, **kwargs)

    def run_checks(self, ws):
        pass

################################################################################
# Combined surface
################################################################################

class CombinedSurface(ArtsObject):
    """
    This class combines two surface models and switches between them depending
    on a surface property variable provided by the data provider.
    """
    def __init__(self,
                 surface_model_1,
                 surface_model_2,
                 surface_type_variable = "surface_type"):
        """
        Args:
            surface_model_1: The surface model to use when surface type is 0
            surface_model_2: The surface model to use when surface type is 1
            surface_type_variable: The variable based on which to switch between
                 the two models.
        """
        super().__init__()
        self.surface_model_1 = surface_model_1
        self.surface_model_2 = surface_model_2
        self.surface_type_variable = surface_type_variable

    @property
    def required_data(self):
        return self.surface_model_1.required_data + self.surface_model_2.required_data

    @arts_property(group="Index")
    def surface_type(self):
        return None

    @property
    def surface_agenda(self):

        @arts_agenda
        def surface_agenda(ws):
            ws.Ignore(ws.f_grid)
            ws.Ignore(ws.rtp_pos)
            ws.Ignore(ws.rtp_los)
            ws.Touch(ws.surface_skin_t)
            ws.Touch(ws.surface_emission)
            ws.Touch(ws.surface_los)
            ws.Touch(ws.surface_rmatrix)
            if (self.surface_type == 0):
                ws.execute_agenda(self.surface_agenda_1)
            else:
                ws.execute_agenda(self.surface_agenda_2)

        return surface_agenda

    def setup(self, ws):
        self.surface_model_1.setup(ws)
        self.surface_model_2.setup(ws)
        ws.Copy(ws.surface_rtprop_agenda, self.surface_agenda)


    def get_data(self, ws, data_provider, *args, **kwargs):
        self.surface_model_1.get_data(ws, data_provider, *args, **kwargs)
        self.surface_model_2.get_data(ws, data_provider, *args, **kwargs)
        self.surface_agenda_1 = self.surface_model_1.surface_agenda
        self.surface_agenda_2 = self.surface_model_2.surface_agenda
        self.get_data_arts_properties(ws, data_provider, *args, **kwargs)

        if self.surface_type <= 0.0:
            ws.Copy(ws.surface_rtprop_agenda, self.surface_model_1.surface_agenda)
        else:
            ws.Copy(ws.surface_rtprop_agenda, self.surface_model_2.surface_agenda)
        ws.Copy(ws.surface_rtprop_agenda, self.surface_agenda)

    def run_checks(self, ws):
        pass



"""
artssat.arts_object
-------------------

The :code:`arts_object` module provides meta-programming functionality
that aims to simplify unified treatment of relevant ARTS variables.
The basic idea is to provide so called *ARTS properties*, which are
similar in concept to Python properties but extend their functionality to
simplify handling of variables that should be passed to the ARTS workspace.

Concept
=======

From the user perspective, the purpose of ARTS properties is to provide
a simple interface to set ARTS parameters using attributes of Python
objects, hiding away boilerplate code related to the management of
these parameters. From the developer perspective they provide a unified
way to treat these variables as well as default methods for the
interaction with the ARTS workspace. The implementation of ARTS properties
is based on the
`Descriptor <https://docs.python.org/3.7/howto/descriptor.html>`_ protocol.

Usecase
~~~~~~~

The typical usage scenario is that a Python class is used to represent
a conceptual unit grouping together various ARTS functionality and
variables. An example of this is for example the :class:`Sensor` class.
In ARTS, the sensor is described through a set of WSVs. ARTS properties
can then for example be used to expose the :code:`f_grid` WSV as an
attributes of the :code:`Sensor` class.

::

    class Sensor:
        @arts_property
        def f_grid(group = "Vector", shape = (Dimension.Frq), wsv = f_grid):
             return None # No default value

User perspective
~~~~~~~~~~~~~~~~

Seen from the user perspective, ARTS properties provide the following
functionality:

1. Provision of default values
2. A unified protocol for the setting of simulation parameters

ARTS properties can be set using Python assignment syntax, i.e.

::

    sensor.f_grid = np.array([1.0, 2.0, 3.0])

or they may be obtained from the :code:`data_provider` using the
appropriate get method when a concrete simulation is run. ARTS
properties impelement the required functionality to provide this
functionality to the user in an opaque manner.

Developer perspective
~~~~~~~~~~~~~~~~~~~~~

For the developer, the purpose of ARTS properties is to provide
meta-programming functionality to simplify the handling of ARTS-related
variables. It provides default methods for the setting of ARTS properties
as well as default :code:`setup` and :code:`get_data` methods.

Additionally, the :code:`arts_object` provides functionality for
the propagation of dimension information throughout :code:`artssat`. This
is important to enable testing consistency of user-provided data.

Reference
=========
"""
import numpy as np
import inspect
from abc import ABCMeta, abstractmethod, abstractproperty
from copy import copy
from pyarts.workspace.variables import WorkspaceVariable, group_names
from pyarts.workspace.variables import workspace_variables as wsv
from pyarts.workspace import Workspace

################################################################################
# The Dimension class
################################################################################

class Dimension:
    """
    Service class to keep track of related dimensions throughout an ARTS
    simulation.

    A consistent ARTS simulation setup requires provided data to have the
    same sizes along related dimensions. The :code:`Dimensions` class
    provides a symbolic representation of these dimensions and provides
    functionality to propagate information on the values of these dimensions
    as data is provided by the user.
    """

    #
    # Singleton classes describing dimensions
    #
    class P:
        """Singleton class representing the pressure dimension."""
        @classmethod
        def __repr__(self):
            return "pressure"

    class Lat:
        """Singleton class representing the latitude dimension."""
        @classmethod
        def __repr__(self):
            return "latitude"

    class Lon:
        """Singleton class representing the longitude dimension."""
        @classmethod
        def __repr__(self):
            return "longitude"

    class Atm:
        """
        Singleton class representing the number of dimensions of the
        Atmosphere.
        """
        @classmethod
        def __repr__(self):
            return "atmosphere"

    class Los:
        """ Singleton class representing the line-of-sight dimension. """
        @classmethod
        def __repr__(self):
            return "line-of-sight"

    class Obs:
        """
        Singleton dimension representing the number of measurement
        blocks.
        """
        @classmethod
        def __repr__(cls):
            return "measurement block"

    class Frq:
        """Singleton class representing the number of frequencies."""
        @classmethod
        def __repr__(self):
            return "frequency"

    class Joker:
        """
        Singleton class representing dimensions that can take an
        arbitrary value.
        """
        pass

    dimensions = [P, Lat, Lon, Atm, Los, Obs, Frq, Joker]

    def __init__(self):
        """
        Create a dimensions object.
        """
        self.dimensions = {}

    def infer(self, dim):
        """
        Infer the value of a dimension.

        Parameters:

            dim: Singleton class representing the dimension to lookup.

        Returns:

            None if no value has been deduced for the requested dimension.

            Otherwise a tuple :code:`(value, who)` consisting of an integer
            value :code:`value` representing the deduced value of the requested
            dimension as well as a :code:`string` :code:`who` indicating
            the variable from which the dimensions has been deduced
        """
        if dim is Dimension.Joker:
            return None

        if dim in self.dimensions:
            return self.dimensions[dim]
        else:
            return None

    def deduce(self, dim, value, who):
        """
        Add a deduction of a value of a given dimension.

        This function should be called when data has been provided by the
        user that allows the deduction of the size of a given dimension
        if no previous such value has been deduced, i.e. the :code:`infer`
        method for the same dimension returns :code:`None`

        Parameters:

            dim: Singleton class representing the dimension to lookup.

            value(int): The value of the dimension that could be deduced.

            who(str): Name of the variable from which the value has been
                deduced.

        """
        if dim is Dimension.Joker:
            return None

        if dim in self.dimensions:
            other, _ = self.dimensions[dim]
            if not other == value:
                raise Exception("A conflicting value for dimension {0} has "\
                                "aready been deduced.".format(dim))
        else:
            self.dimensions[dim] = (value, who)

    def link(self, other):
        """
        Link to :code:`Dimension` objects.

        This registers the :code:`self` :code:`Dimension` object as master of the
        :code:`other` :code:`Dimensions` object. Deduced dimensions values from
        the *master* dimension are propagated to the *slave* dimensions but not
        vice versa. This allows having dimensions with different values in the
        *slave* objects while still ensuring consistency with the *master* object.

        Parameters:

            other(:code:`Dimensions`): The *slave* dimension to link to this
            dimension.

        """
        # Check consistency of dimension objects.
        for k in self.dimensions:
            if k in other.dimensions:
                v1, who1 = self.dimensions[k]
                v2, who2 = other.dimensions[k]

                if v1 == v2:
                    continue

                if v1 == 1 or v2 == 1:
                    continue

                s = r"Contradictory dimensions deduced for dimension {0}:" \
                    "{1} deduced from {2} and {3} deduced from {4}."\
                    .format(k, v1, who1, v2, who2)
                raise Exception(s)
            else:
                other.dimensions[k] = self.dimensions[k]

        self.dimensions = other.dimensions

################################################################################
# ARTS properties
################################################################################

def get_shape(obj):
    """
    Get the shape of a :code:'numpy.ndarray' or of a nested list.

    Parameters(obj):
       obj: The object of which to determine the shape.

    Returns:

        A tuple describing the shape of the :code:`ndarray` or the
        nested list or :code:`(1,)`` if obj is not an instance of either
        of these types.

    """
    if hasattr(obj, "shape"):
        return obj.shape
    elif type(obj) == list:
        if obj == []:
            return (0,)
        else:
            return (len(obj),) + get_shape(obj[0])
    else:
        return ()

def broadcast(shape, obj):
    """
    Broadcast an array or a list to a provided shape.

    Parameters:

        shape(tuple): Tuple of :code:`int` describing the shape to broadcast
        the provided object to.

        obj(object): The object to broadcast to the provided shape. Either a
        numpy array or a (nested) list.
    """

    if shape == get_shape(obj):
        return obj

    if type(obj) == np.ndarray:
        return np.broadcast_to(obj, shape)
    elif type(obj) == list:
        if len(obj) == 1:
            return shape[0] * [broadcast(shape[1:], obj[0])]
        else:
            return [broadcast(shape[1:], nested) for nested in obj]
    else:
        if shape == ():
            return obj
        else:
            raise Exception("Degree of nesting of list is inconsitent with"\
                            " length of the provided shape tuple.")


def arts_property(group, shape = None, wsv = None, optional = False):
    """
    The :code:`arts_property` decorator.

    This decorator turns a function defintion into an ARTS property. Its useage
    is similar to the :code:`@property` decorator in Python. The ARTS property
    decorator provides default getter, setter, :code:`setup` and :code:`get_data`
    methods. The setter as well as the :code:`setup` and :code:`get_data` methods
    can be overwritten to allow specializing the handling of specific variables.
    For more details see the :class:`ArtsProperty` class.
    ::
        @arts_property
        def f_grid:
            return np.array([187e9]) # The default value

        @f_grid.setter
        def set_f_grid(self, value):
            print("Setting the f_grid.")
            self.value = value

        @f_grid.setter
        def set_f_grid(self, value):
            print("Setting the f_grid.")
            self.value = value

        @f_grid.setup
        def setup_f_grid(self, value):
            print("Customg setup method.")

        @f_grid.get_data
        def get_data_f_grid(self, value):
            print("Customg setup method.")

    Parameters:

        group(str): The ARTS group to which the variable should belong.

        shape(tuple): Tuple describing the expected shape of the variable.

        wsv(pyarts.workspace.WorkspaceVariable): The workspace variable
        corresponding to the ARTS property.
    """
    class ArtsPropertySpecialization(ArtsProperty):
        def __init__(self, fdefault):
            super().__init__(fdefault, group, shape, wsv, optional)
    return ArtsPropertySpecialization

ws = Workspace(verbosity = 0)

class ArtsProperty:
    """
    The :code:`ArtsProperty` class that implements the main functionality of
    ARTS properties.

    An ARTS property is an abstract representation of a parameter of an ARTS
    simulation. This parameter can be set to a fixed value by the user during
    the setting up of the simulation. In this case the property is said to
    be *fixed*. Otherwise, and if no default value for the parameter is set,
    the value will be requested from the data provider.

    The :code: `ArtsProperty` class implements the follwing functionality:

        1. Provide a default setter, that checks consistency with expected ARTS
        group and shape.

        2. Provide a default setup method that sets the value of the
        corresponding ARTS WSV if the value of the property is set to fixed.


    The :code:`ArtsProperty` class is to be used through the :code:
    `arts_property` decorator, which returns a subclass of this class that
    fixes the :code:`group, shape` and :code:`wsv` values of the :code:`__init__`
    method.
    """
    def __init__(self, fdefault, group, shape, wsv, optional):
        """
        Create a :code:`ArtsProperty` instance.

        Parameters:

            fdefault: Default value for the ARTS Property or :code:`None` if no
            reasonable default value can be provided.

            group(str): Name of the ARTS group the value belongs to.

            shape(tuple): Tuple describing the expected shape of the variable.
            Set to :code:`None` if no reasonable expected shape can be
            specified.

            wsv(pyarts.workspace.WorkspaceVariable): Workspace variable
            corresponding to this ARTS property or :code:`None` if no such
            WSV exists.

            optional(Boolean): If True no Exception will be thrown if the
            data_provider doesn't provide a get method for this property.
        """
        self.group = group
        self.shape = shape
        self.optional = optional

        if not wsv is None:
            if type(wsv) == str:
                if hasattr(ws, wsv):
                    wsv = getattr(ws, wsv)
                else:
                    raise Exception("Workspace variable {0} associated with"\
                                    " ARTS property {1} does not exist."\
                                    .format(wsv, fdefault.__name__))
        self.wsv = wsv
        self.fsetup    = None
        self.fget_data = None
        self.fset      = None

        self.name  = fdefault.__name__
        self.fdefault = fdefault

    def get_name(self, owner, separator = "_"):
        """
        Return a qualified name of the ARTS property.

        This return a name of the ARTS property prefixed by the name of the
        owner is this object possesses a :code:`name` attribute.

        Parameters:

            owner: The object instance that this ARTS property belongs
            to.

            separator(str): Char to use to separater the owners name and
            the name of the ARTS property.

        """
        if hasattr(owner, "name"):
            name = getattr(owner, "name") + separator
        else:
            name = ""
        name += self.name
        return name

    def setup(self, fsetup):
        """
        Decorator to set the setup method of the ARTS property. The expected
        call signature of the setup method is :code:`setup(self, obj, ws)`
        where :code:`obj` is the owner of the ARTS property. Providing of the
        :code:`obj` argument of the :code:`setup` method is currently necessary
        to allow the checking of dimensions.

        Parameters:

            fsetup(function): The customized setup function to use for this
            ARTS property.

        """
        self.fsetup = setup

    def get_data(self, fget_data):
        """
        Decorator to set the :code:`get_data` method of the ARTS property.
        The expected call signature of the :code:`get_data` method is
        :code:`get(self, obj, ws, *args, **args)` where :code:`obj` is
        the owner of the ARTS property. Providing of the :code:`obj` argument
        of the :code:`setup` method is currently necessary to allow the checking
        of dimensions.

        Parameters:

            fsetup(function): The customized :code:`get_data` method to use for
            this ARTS property.

        """
        self.fget_data = fget_data

    def setter(self, fset):
        """
        Decorator to set the setter to set the value of this ARTS property.
        The expected call signature is :code:`set(self, value)` and should
        set the :code:`value` attribute of the :code:`ArtsProperty` object
        :code:`self`.

        Parameters:

            fset(function): The customized setter to use for this
            :code:`ArtsPorperty`

        """
        self.fset = fset

    def check_and_convert(self, value):
        """
        Checks type of :code:`value` against the group specification of this
        ARTS property contained in :code:`self.group`.

        The group specification can be a single string in which case this
        function will try to convert the provided value to the given group
        using the :code:`convert` class method of
        :class:`pyarts.workspace.WorkspaceVariable`.

        It is also possible to specify a list of groups for the expected
        value. In this case this function simply checks whether the group
        inferred by :code:`WorkspaceVariable.get_group_id`
        group of the :code:``

        """
        if type(self.group) == str:
            converted = WorkspaceVariable.convert(self.group, value)

            if converted is None:
                raise Exception("Provided value of type {0} cannot be converted"
                                " to ARTS group {1}".format(type(value),
                                                            self.group))
            value = converted
        elif type(self.group) == list:
            g_i = WorkspaceVariable.get_group_id(value)
            g = group_names[g_i]
            if not g in self.group:
                raise Exception("Provided value of type {0} is not of any of "
                                " the expected ARTS groups {1}."\
                                .format(type(value), self.group))

        return value

    def check_and_broadcast(self, value, owner):
        """
        check the shape of a given variable against a symbolic shape specification.

        Parameters:

            value(array or list): the object of which to check the shape.

            who(str): name of the variable of which the shape is checked.

        Raises:

            Exception if the shape of :code:`value` is inconsitent with the shape
            specification the ARTS property.

        """
        who = self.get_name(owner, separator = ".")
        shape = get_shape(value)

        if self.shape is None:
            return value

        # catch inconsistent number of dimensions.
        if not len(shape) == len(self.shape):

            s = "The provided value for {0} has {1} dimensions but {2} were"\
                " expected.".format(who, len(shape), len(self.shape))
            raise Exception(s)

        # deduce and compare dimensions.
        error = "The provided value has dimension {1} along axis {0} but {2}" \
                "were expected."

        deduced = tuple()
        for i in range(len(self.shape)):

            # Fixed dimension
            if type(self.shape[i]) == int:
                deduced += (shape[i],)
                if shape[i] == self.shape[i]:
                    continue
                raise Exception(error.format(i, shape[i], self.shape[i]))

            # Symbolic dimension
            elif self.shape[i] in Dimension.dimensions:
                d = owner.dimensions.infer(self.shape[i])
                if d is None:
                    owner.dimensions.deduce(self.shape[i], shape[i], who)
                    deduced += (shape[i],)
                else:
                    n, who2 = d
                    deduced += (n,)
                    if n == shape[i] or shape[i] == 1:
                        continue
                    s = "The value provided for the {0} property was expected"\
                        " to match the {1} dimension of the simulation  along"\
                        " axis {2} but this is not the case. The value of the"\
                        " {1} dimension has been deduced to be to be {3}"\
                        " from the value of the {4} property."
                    raise Exception(s.format(who, self.shape[i], i, n, who2))
            else:
                raise Exception("Shape specification should consist of either"\
                                " integers or symbolic dimensions.")

        value = broadcast(deduced, value)
        return value

    def __get__(self, owner, objtype = None):

        if owner is None:
            return self

        val = owner.__dict__["_" + self.name].value

        if val is None:
            val = self.fdefault(owner)

        return val

    def __set__(self, owner, value):
        if not self.fset is None:
            self.fset(owner, value)
            return None

        value = self.check_and_convert(value)

        if not self.shape is None:
            value = self.check_and_broadcast(value, owner)

        ph = owner.__dict__["_" + self.name]
        ph.fixed = True
        ph.value = value

        if not self.wsv is None:
            if not ph.workspace is None:
                owner.set_wsv(ph.workspace, self.wsv, value)

    def _setup(self, *args, **kwargs):
        """
        Run the default or customized :code:`setup` method of the
        :code:`ArtsProperty` object.
        """
        if self.fsetup is None:
            self._default_setup(*args)
        else:
            self.fsetup(*args, **kwargs)

    def _get_data(self, owner, ws, data_provider, *args, **kwargs):
        """
        Run the default or customized :code:`get_data` method of the
        :code:`ArtsProperty` object.
        """
        if self.fget_data is None:
            self._default_get_data(owner, ws, data_provider,
                                   *args, **kwargs)
        else:
            self.fget_data(owner, ws, data_provider,
                           *args, **kwargs)

    def _default_setup(self, owner, ws):
        """
        The default :code:`setup` method of the :code:`ArtsProperty`.
        This method simply checked whether the value of the :code:`ArtsProperty`
        was set to a fixed value. If this is the case and the :code:`wsv`
        attribute contains an associated WSV, the WSVs value in the
        given workspace :code:`ws` is set.

        Parameters:

            owner(obj): The object that this :code:`ArtsProperty` belongs
            to.

            ws(arts.typhon.workspace.Workspace): Ths workspace which to
            setup.
        """
        ph = owner.__dict__["_" + self.name]
        ph.workspace = ws
        if self.wsv and ph.fixed:
            owner.set_wsv(ws, self.wsv, ph.value)

    def _default_get_data(self, owner, ws, data_provider, *args, **kwargs):
        """
        The default :code:`get_data` method. It performs the following steps:

            1. Check if the :code:`data_provider` provides a get method for the
            :code:`ArtsProperty`. If so, then set the associated WSV to this
            value (if given).

            2. Check if the value of the :code:`ArtsProperty` is fixed. In this
            case it has been set already during setup, so no additional action
            is required.

            3. Check if a default value has been provided for the
            :code:`ArtsProperty`. In this case use this value.

            4. Throw an exception if all of the previous steps were unsuccessful
            in determining the value of the :code:`ArtsProperty`.
        """
        ph = owner.__dict__["_" + self.name]
        if not ph.fixed:
            getter_name = "get_" + self.get_name(owner, separator = "_")

            default = self.fdefault(owner)

            # Try to get value from provider.
            if hasattr(data_provider, getter_name):
                f = getattr(data_provider, getter_name)
                value = f(*args, **kwargs)

                value = self.check_and_convert(value)

                if not self.shape is None:
                    value = self.check_and_broadcast(value, owner)

                if self.wsv:
                    owner.set_wsv(ws, self.wsv, value)

                ph = owner.__dict__["_" + self.name]
                ph.value = value

            # Check if there's a default value.
            elif not default is None:
                ph = owner.__dict__["_" + self.name]
                ph.value = default

                if self.wsv:
                    owner.set_wsv(ws, self.wsv, default)

            # No value - throw exception
            elif not self.optional:
                raise Exception("Neither a default value nor a get method "
                                " has been provided for the ARTS property "
                                "{0}.".format(self.get_name(owner, ".")))
            else:
                return


class PlaceHolder:
    """
    Data that is required for an ARTS simulation and can be set
    either to a fixed value, or taken from a data provider.
    """
    def __init__(self):
        """
        Create a PlaceHolder object representing the property
        :code:`name` with expected dimensions :code:`dimensions` and
        expected type :code:`expected_type`.

        Parameters:

            name(str): The name of the property for which this object
                is the placeholder.
            dimensions(tuple): A tuple of :code:`Dimension` objects
                describing the expected dimensions of the corresponding
                property.
            expected_type(type): The expected type for the corresponding
                property.
        """
        self.fixed = False
        self.value = None
        self.workspace = None

    def __getstate__(self):
        state = copy(self.__dict__)
        state.pop("workspace", None)
        return state

def add_property(obj, name, dims, t):
    """
    Add an ARTS property to an existing object.

    Parameters:

        obj(:code:`object`): The object to add the property to.

        name(:code: `str`): Name of the property to create

        dims(:code:`tuple`): Tuple describing the expected dimensions
        of the property.

        t(:code:`type`): The expected type of the property.
    """
    getter = make_getter(name)
    setter = make_setter(name)
    prop = property(getter, setter, name)
    setattr(type(obj), name, prop)

    ph = PlaceHolder(name, dims, t)
    setattr(obj, "_" + name, ph)

def is_arts_property(obj):
    return isinstance(obj, ArtsProperty)

class ArtsObject:
    """
    The :ArtsObject: class provides a base class for objects that bundle
    ARTS workspace variables and functionality into a conceptual unit. The
    class provides service functions that automate the handling of ARTS
    properties.
    """

    def __init__(self):
        """
        Create an :code:`ArtsObject` instance.
        """
        self._wsvs      = {}
        self.dimensions = Dimension()

        for _, ap in inspect.getmembers(type(self), is_arts_property):
            self.__dict__["_" + ap.name] = PlaceHolder()

    def setup_arts_properties(self, ws):
        """
        Run the :code:`setup` method for all ARTS properties of the class.

        Parameters:

            ws(pyarts.workspace): The workspace for which to setup
            the simulation.

        """
        for _, ap in inspect.getmembers(type(self), is_arts_property):
            ap._setup(self, ws)

    def setup(self, ws):
        self.setup_arts_properties(ws)

    def get_data_arts_properties(self, ws, data_provider, *args, **kwargs):
        """
        Run the :code:`get_data` method for all ARTS properties of this
        object.

        Parameters:

            ws(pyarts.workspace): The workspace for which to setup
            the simulation.

            data_provider(obj): The :code:`data_provider` providing the
            data for the simulation.

            *args: Additional parameters to be passed on to the data
            provider.

            **kwargs: Additional keyword arguments to be passed to the data
            provider.
        """
        for _, ap in inspect.getmembers(type(self), is_arts_property):
            ap._get_data(self, ws, data_provider, *args, **kwargs)

    def get_data(self, ws, data_provider, *args, **kwargs):
        self.get_data_arts_properties(ws, data_provider, *args, **kwargs)

    def set_wsv(self, ws, wsv, value):
        """
        Sets the given private WSV :code:`wsv` of the object or if the
        object doesn't have a private copy of :code:`wsv` then sets :code:`wsv`
        on the workspace :code:`ws`.

        Parameters:

            ws(pyarts.workspace.Workspace): The workspace in which to set
            the WSV.

            wsv(pyarts.workspace.variables.WorkspaceVariable): The
            variable to set.

            value(obj): The value to set the WSV :code:`wsv` to.
        """
        if wsv.name in self._wsvs:
            self._wsvs[wsv.name].ws    = ws
            self._wsvs[wsv.name].value = value
        else:
            setattr(ws, wsv.name, value)

    def update_wsv(self, wsv, value):
        """
        Updates the value of a given WSV considering private WSVs of the owner.
        This requires that the variable has already been set to a value during
        so that is contains a reference to the workspace in which it is used.

        Parameters:

            wsv(pyarts.workspace.variables.WorkspaceVariable): The
            variable to set.

            value(obj): The value to set the WSV :code:`wsv` to.
        """
        if wsv.name in self._wsvs:
            self._wsvs[wsv.name].value = value
        else:
            wsv.value = value

    def call_wsm(self, ws, wsm):
        """
        Call workspace method on the given workspace.

        This method replaces inputs of the workspace variable with the private
        WSMs of the object. After execution of the method the results are
        copied to the private WSMs of the object.

        Parameters:

            ws(pyarts.workspace.Workspace): The workspace on which to
            execute the method.

            wsm(pyarts.workspace.WorkspaceMethod): The workspace method
            to execute.

        """
        args = self.get_wsm_kwargs(wsm)
        wsm.call(ws, **args)

        # Copy output
        for i in wsm.outs:
            name = WorkspaceVariable.get_variable_name(i)
            if name in self._wsvs and not i in wsm.ins:
                ws.Copy(self._wsvs[name], wsv[name])

    def _create_private_wsvs(self, ws, names):
        """
        Create private copies of given WSV names.

        Parameters:

            ws(:code:`pyarts.workspace.Workspace`): A workspace instance
            on which to create the workspace variables.

            names(list): List of strings containing the names of the workspace
            variables to create.
        """
        for name in names:
            wsv = ws.__getattr__(name)
            wsv_private = ws.create_variable(wsv.group,
                                             self.name + "_" + name)
            self._wsvs[name] = wsv_private

    def get_wsm_kwargs(self, wsm):
        """
        Generate a list of arguments to the given ARTS workspace method
        :code:`wsm` for which the sensor related input parameters are
        replace by the ones of this sensor. This is done by checking
        whether the input argument name is in the sensors :code:`_wsv`
        dictionary and if so replacing the argument.

        Parameters:

           wsm(pyarts.workspace.methods.WorkspaceMethod): The ARTS
               workspace method object for which to generate the input
               argument list.

        Returns:

            The list of input arguments with sensor specific input arguments
            replaced by the corresponding WSVs of the sensor.

        """
        kwargs = {}
        for i in wsm.ins:
            name = WorkspaceVariable.get_variable_name(i)
            if name in self._wsvs:
                kwargs[name] = self._wsvs[name]
            else:
                kwargs[name] = wsv[name]
        return kwargs






"""
artssat.solvers
===============

This module contains scattering solver classes that represent the
different scattering solver class that are available in ARTS.

It provides the abstract :class:`ScatteringSolver` class which defines
the general scattering solver interface as well as concrete implementations
of solvers available in ARTS.

Reference
=========
"""
import numpy as np
from abc import ABCMeta, abstractproperty, abstractmethod
from pyarts.workspace.methods import workspace_methods
from pyarts.workspace.variables import workspace_variables

wsv = workspace_variables
wsm = workspace_methods

################################################################################
# Scattering solver
################################################################################

class ScatteringSolver(metaclass = ABCMeta):
    """
    Abstract base class that defines the scattering solver interface.
    """
    def __init__(self):
        pass

    @abstractmethod
    def make_solver_call(self, sensor):
        """
        This method should return a function :code:`run(ws)` that runs the
        scattering solver on a given workspace :code:`ws`. This function is
        called for each sensor before the corresponding pencil beam calculations
        are performed. To be used within a retrieval this method must be
        convertible to an ARTS agenda.
        """
        pass

################################################################################
# RT4 solver
################################################################################

class RT4(ScatteringSolver):
    def __init__(self,
                 nstreams = 32,
                 pfct_method = "median",
                 quad_type = "D",
                 add_straight_angles = 1,
                 pfct_aa_grid_size = 38,
                 auto_inc_nstreams = 64,
                 robust = 1):
        self._nstreams = nstreams
        self._pfct_method = pfct_method
        self._quad_type = quad_type
        self._add_straight_angles = add_straight_angles
        self._pfct_aa_grid_size = pfct_aa_grid_size
        self._auto_inc_nstreams = auto_inc_nstreams
        self._robust = robust

    #
    # Properties
    #

    @property
    def nstreams(self):
        return self._nstreams

    @property
    def pfct_method(self):
        return self._pfct_method

    @property
    def quad_type(self):
        return self._quad_type

    @property
    def add_straight_angles(self):
        return self._add_straight_angles

    @property
    def pfct_aa_grid_size(self):
        return self._pfct_aa_grid_size

    @property
    def auto_inc_nstreams(self):
        return self._auto_inc_nstreams

    @property
    def robust(self):
        return self._robust

    def make_solver_call(self, atmosphere, sensor):

        kwargs = sensor.get_wsm_kwargs(wsm["RT4Calc"])
        def run_solver(ws):
            ws.RT4Calc(**kwargs,
                       nstreams = self._nstreams,
                       pfct_method = self._pfct_method,
                       quad_type = self._quad_type,
                       add_straight_angles = self._add_straight_angles,
                       pfct_aa_grid_size = self._pfct_aa_grid_size,
                       auto_inc_nstreams = self._auto_inc_nstreams,
                       robust = self.robust)

        return run_solver

class Disort(ScatteringSolver):

    def __init__(self,
                 nstreams = 8,
                 pfct_method = "interpolate",
                 new_optprop = 1,
                 Npfct = 181):

        self._nstreams = nstreams
        self._pfct_method = pfct_method
        self._new_optprop = new_optprop
        self._Npfct = Npfct

    def make_solver_call(self, atmosphere, sensor):

        kwargs = sensor.get_wsm_kwargs(wsm["DisortCalcWithARTSSurface"])

        def run_solver(ws):
            ws.Ignore(ws.atmosphere_dim)
            ws.DOAngularGridsSet(N_za_grid = 38, N_aa_grid = 1, za_grid_opt_file = "")

            ws.DisortCalcWithARTSSurface(**kwargs,
                                         nstreams = self._nstreams,
                                         pfct_method = self._pfct_method,
                                         Npfct = self._Npfct)

        return run_solver




import os
import glob

if "SSDB_PATH" in os.environ:
    ssdb_path = os.environ["SSDB_PATH"]
else:
    ssdb_path = "/home/simonpf/src/jupyterhub/data/scattering"

class Shape:
    def __init__(self, name, type = "ice"):
        if type == "ice":
            self.type = "Ice"
        else:
            self.type = "Liquid"

        self.name = name
        #self.path = os.path.join(ssdb_path, "StandardHabits", "FullSet", name + ".xml")
        #self.meta = os.path.join(ssdb_path, "StandardHabits", "FullSet", name + ".meta.xml")
        self.path = os.path.join(ssdb_path, name, name + ".xml")
        self.meta = os.path.join(ssdb_path, name, name + ".meta.xml")


        self.img = os.path.join(ssdb_path, name, "shape_img.png")
        #self.img  = glob.glob(os.path.join(ssdb_path, "SSD", "TotallyRandom", "*", "*", "*",
        #                                   "*" + name.replace("-", "") + "*",
        #                                   "shape_img.png"))[0]

    def copy_to(self, path):
        from shutil import copy
        habit_path = os.path.join(path, self.name)
        if not os.path.exists(habit_path):
            os.mkdir(habit_path)

        copy(self.path, habit_path)
        copy(self.path + ".bin", habit_path)
        copy(self.meta, habit_path)
        copy(self.img, habit_path)

    def __repr__(self):
        return self.name

#standard_habits = glob.glob(os.path.join(ssdb_path, "StandardHabits", "FullSet", "*"))

#names = ["LargePlateAggregate", "6-BulletRosette", "EvansSnowAggregate", "LiquidSphere"]
#for n in names:
#    s = Shape(n)
#    s.copy_to("/home/simonpf/src/jupyterhub/data/scattering")
names = glob.glob(os.path.join(ssdb_path, "*"))
shapes = []
for n in names:
    s = Shape(os.path.basename(n).split(".")[0])
    shapes += [s]





import numpy as np
from artssat.atmosphere.atmospheric_quantity \
    import AtmosphericQuantity, extend_dimensions
from artssat.arts_object import add_property
from artssat.scattering.psd.arts.arts_psd import ArtsPSD

from artssat.jacobian  import JacobianBase
from artssat.retrieval import RetrievalBase, RetrievalQuantity

from pyarts.workspace import arts_agenda

################################################################################
# Jacobian
################################################################################

class Jacobian(JacobianBase):

    def __init__(self, quantity, index):
        super().__init__(quantity, index)


    def _make_setup_kwargs(self, ws):

        kwargs = self.get_grids(ws)
        kwargs.update({"species" : self.quantity._species_name,
                      "quantity" : self.quantity.name})

        return kwargs

    def setup(self, ws):

        kwargs = self._make_setup_kwargs(ws)
        ws.jacobianAddScatSpecies(**kwargs)

################################################################################
# Retrieval
################################################################################

class Retrieval(RetrievalBase, Jacobian):

    def __init__(self, quantity, index):
        RetrievalBase.__init__(self)
        Jacobian.__init__(self, quantity, index)

    def add(self, ws):
        ws.retrievalAddScatSpecies(**self._make_setup_kwargs(ws))

class Moment(AtmosphericQuantity, RetrievalQuantity):
    def __init__(self,
                 species_name,
                 moment_name,
                 data = None,
                 jacobian = False):

        name = species_name + "_" + moment_name
        AtmosphericQuantity.__init__(self, name, (0, 0, 0), jacobian)
        RetrievalQuantity.__init__(self)


        self._jacobian = None
        self._retrieval = None
        self._species_name = species_name
        self._moment_name  = moment_name
        self._data = data

    #
    # Abstract methods
    #

    def setup(self, ws, i):
        self._wsv = ws.particle_bulkprop_field
        self._wsv_index = i

    def setup_jacobian():
        pass

    def get_data(self, ws, provider, *args, **kwargs):

        if self.retrieval is None:

            if hasattr(provider, "get_" + self.name):
                AtmosphericQuantity.get_data(self, ws, provider, *args, **kwargs)
            else:
                try:
                    x = self.data
                    pbf_shape = ws.particle_bulkprop_field.value.shape[1:]
                    if len(x.shape) < len(pbf_shape):
                        n = len(pbf_shape) - len(x.shape)
                        x = np.reshape(x, x.shape + (1,) * n)
                    ws.particle_bulkprop_field.value[self._wsv_index, :, :, :] \
                        = np.broadcast_to(x, pbf_shape)
                except Exception as e:
                    raise Exception("Encountered error trying to get data for "
                                    " moment {0}: {1}".format(self.name, e))

    #
    # Jacobian & retrieval
    #

    @property
    def jacobian_class(self):
        return Jacobian

    @property
    def retrieval_class(self):
        return Retrieval

    def set_from_x(self, ws, x):

        if not self.retrieval.limit_high == None:
            x = np.minimum(x, self.retrieval.limit_high)

        if not self.retrieval.limit_low == None:
            x = np.maximum(x, self.retrieval.limit_low)

        grids = [ws.p_grid.value, ws.lat_grid.value, ws.lon_grid.value]
        grids = [g for g in grids if g.size > 0]

        x = self.transformation.invert(x)
        x = self.retrieval.interpolate_to_grids(x, grids)

        pbf_shape = ws.particle_bulkprop_field.value.shape[1:]
        ws.particle_bulkprop_field.value[self._wsv_index, :, :, :] = np.reshape(x, pbf_shape)

    #
    # Properties
    #

    @property
    def data(self):
        return self._data

    def species_name(self):
        return self._species_name

class ScatteringSpecies:
    def __init__(self,
                 name,
                 psd,
                 scattering_data,
                 scattering_meta_data = None,
                 jacobian = False):

        self._name = name
        self._psd  = psd

        if hasattr(scattering_data, "path") and hasattr(scattering_data, "meta"):
            scattering_meta_data = scattering_data.meta
            scattering_data = scattering_data.path

        if not scattering_data[-4:] in [".xml", "l.gz"]:
            scattering_data += ".xml"

        self._scattering_data = scattering_data

        if scattering_meta_data is None:
            md = scattering_data[:-3] + "meta.xml"
            self._scattering_meta_data = md
        else:
            self._scattering_meta_data = scattering_meta_data

        self._moments = []

        try:
            moment_data = self.psd.moments
            for m, d in zip(self.psd.moment_names, moment_data):
                moment = Moment(self.name, m, data = d)
                self._moments += [moment]
                self.__dict__[m] = moment
        except:
            for m in self.psd.moment_names:
                moment = Moment(self.name, m)
                self._moments += [moment]
                self.__dict__[m] = moment

    @property
    def name(self):
        return self._name

    @property
    def moment_names(self):
        name = self._name
        return [name + "_" + m for m in self._psd.moment_names]

    @property
    def psd(self):
        return self._psd

    @psd.setter
    def psd(self, psd):
        if not isinstance(psd, ArtsPSD):
            raise ValueError("PSD of scattering species must implement the ArtsPsd ABC.")
        self._psd = psd

    @property
    def scattering_data(self):
        return self._scattering_data

    @scattering_data.setter
    def scattering_data(self, scattering_data):

        if type(scattering_data) == tuple:
            scattering_data, scattering_meta_data = scattering_data
        elif  hasattr(scattering_data, "path") and hasattr(scattering_data, "meta"):
            scattering_meta_data = scattering_data.meta
            scattering_data = scattering_data.path
        else:
            scattering_meta_data = None

        if not scattering_data[-4:] in [".xml", "l.gz"]:
            scattering_data += ".xml"

        self._scattering_data = scattering_data

        if scattering_meta_data is None:
            md = scattering_data[:-3] + "meta.xml"
            self._scattering_meta_data = md
        else:
            self._scattering_meta_data = scattering_meta_data

    @property
    def scattering_meta_data(self):
        return self._scattering_meta_data

    @property
    def jacobian(self):
        return self._jacobian

    @property
    def moments(self):
        return self._moments

    @property
    def pnd_agenda(self):
        return self._psd.agenda

    def setup(self, ws, i):

        wsv_name = str(id(self))
        ws.ArrayOfSingleScatteringDataCreate(wsv_name)
        ws.ArrayOfScatteringMetaDataCreate(wsv_name + "_meta")
        scat_data_wsv = ws.__getattr__(wsv_name)
        scat_meta_data_wsv = ws.__getattr__(wsv_name + "_meta")
        ws.ReadXML(scat_data_wsv, self.scattering_data)
        ws.ReadXML(scat_meta_data_wsv, self.scattering_meta_data)

        ws.Append(ws.scat_data_raw, scat_data_wsv)
        ws.Append(ws.scat_meta, scat_meta_data_wsv)

        ws.Delete(scat_data_wsv)
        ws.Delete(scat_meta_data_wsv)

        ws.Append(ws.pnd_agenda_array, self.pnd_agenda)
        ws.Append(ws.scat_species, self.name)
        ws.Append(ws.pnd_agenda_array_input_names, self.moment_names)

        for j, m in enumerate(self.moments):
            m.setup(ws, i + j)

        if hasattr(self.psd, "setup"):
            self.psd.setup(ws, i)

    def get_data(self, ws, provider, *args, **kwargs):
        for m in self.moments:
            m.get_data(ws, provider, *args, **kwargs)

        if hasattr(self.psd, "get_data"):
            self.psd.get_data(ws, provider, *args, **kwargs)

    def setup_jacobian(self, ws):
        for m in moments:
            m.setup_jacobian(ws)
